import os
import re
from threading import Thread

from easelenium.ui.generator.page_object_generator import PageObjectGenerator
from easelenium.ui.root_folder import RootFolder
from easelenium.ui.string_utils import StringUtils
from easelenium.ui.utils import FLAG_ALL_AND_EXPAND
from easelenium.ui.widgets.image.selectable_image import SelectableImagePanel
from easelenium.ui.widgets.utils import (
    DialogWithText,
    WxTextCtrlHandler,
    show_dialog,
    show_dialog_bad_name,
    show_dialog_path_does_exist,
    show_dialog_path_doesnt_exist,
)
from easelenium.utils import LINESEP, Logger, get_py_file_name_from_class_name
from selenium.webdriver.common.by import By
from wx import EVT_BUTTON, EVT_MOTION, Button, GridBagSizer, Panel, StaticText, TextCtrl


class GeneratorTab(Panel):
    def __init__(self, parent):
        Panel.__init__(self, parent)
        self.main_frame = self.GetTopLevelParent()

        self.__create_widgets()

    def __create_widgets(self):
        sizer = GridBagSizer(5, 5)

        row = 0
        col = 0

        # first row
        label = StaticText(self, label="Selected area:")
        sizer.Add(label, pos=(row, col))

        col += 1
        self.txt_selected_area = TextCtrl(self, value="(0, 0, 0, 0)")
        sizer.Add(self.txt_selected_area, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        label = StaticText(self, label="Class name:")
        sizer.Add(label, pos=(row, col))

        col += 1
        self.txt_class_name = TextCtrl(self)
        sizer.Add(self.txt_class_name, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.bth_reload_img = Button(self, label="Reload image")
        self.bth_reload_img.Bind(EVT_BUTTON, self.__load_img)
        sizer.Add(self.bth_reload_img, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.btn_generate = Button(self, label="Generate")
        self.btn_generate.Bind(EVT_BUTTON, self.generate)
        sizer.Add(self.btn_generate, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        # second row
        row += 1
        col = 0
        self.select_image_panel = SelectableImagePanel(self)
        self.select_image_panel.static_bitmap.Bind(EVT_MOTION, self._on_mouse_move)
        sizer.Add(
            self.select_image_panel,
            pos=(row, col),
            span=(1, 6),
            flag=FLAG_ALL_AND_EXPAND,
        )

        sizer.AddGrowableCol(1, 1)
        sizer.AddGrowableRow(row, 1)

        self.SetSizer(sizer)

    def __get_root_folder(self):
        return self.main_frame.get_root_folder()

    def __load_img(self, evt=None):
        browser = self.main_frame.get_browser()
        if browser:
            img_path = browser.save_screenshot(self.main_frame.get_tmp_dir())
            self.select_image_panel.load_image(img_path)

            w, h = self.select_image_panel.get_image_dimensions()

            self.txt_selected_area.SetValue("(%d, %d, %d, %d)" % (0, 0, w, h))
            self.main_frame.set_url(browser.get_current_url())

    def _on_mouse_move(self, evt):
        if self.select_image_panel.was_image_loaded():
            SelectableImagePanel.on_mouse_move(self.select_image_panel, evt)
            selected_area = self.select_image_panel.get_selected_area()
            self.txt_selected_area.SetValue(repr(selected_area))

    def __is_gen_data_correct(self):
        root_folder = self.__get_root_folder()
        class_name = self.txt_class_name.GetValue()
        area = self.txt_selected_area.GetValue()
        if not self.main_frame.get_browser():
            msg = "Browser is not opened." + LINESEP + "Please open url."
            caption = "Browser is not opened"
            show_dialog(self, msg, caption)
            return False
        elif not re.match(
            r"\(\s*[0-9]+\s*,\s*[0-9]+\s*,\s*[0-9]+\s*,\s*[0-9]+\s*\)", area
        ):
            msg = "Selected area is not correct: '%s'" % area
            caption = "Bad selected area"
            show_dialog(self, msg, caption)
            return False
        elif root_folder is None or not os.path.exists(root_folder):
            show_dialog_path_doesnt_exist(self, root_folder)
            return False
        elif len(class_name) == 0:  # if bad class_name
            msg = "Unsupported name for class: '%s'" % class_name
            caption = "Bad name for class"
            show_dialog(self, msg, caption)
            return False
        return True

    def __get_frames(self):
        browser = self.main_frame.get_browser()
        if browser:
            frames = []
            for e in browser.find_elements((By.CSS_SELECTOR, "frame, iframe")):
                name = browser.get_attribute(e, "name")
                src = browser.get_attribute(e, "src")
                frames.append((name, src))
            return frames
        else:
            return None

    def generate(self, evt):
        if self.__is_gen_data_correct():
            folder = self.__get_root_folder()
            if RootFolder.PO_FOLDER in os.listdir(folder):
                folder = os.path.join(folder, RootFolder.PO_FOLDER)

            class_name = self.txt_class_name.GetValue()
            file_path = os.path.join(
                folder, get_py_file_name_from_class_name(class_name)
            )
            area_as_text = self.txt_selected_area.GetValue()
            url = self.main_frame.get_url()
            if os.path.exists(file_path):
                show_dialog_path_does_exist(self, file_path)
            elif not StringUtils.is_class_name_correct(class_name):
                show_dialog_bad_name(self, class_name, "Header", "ContextMen")
            elif not StringUtils.is_area_correct(area_as_text):
                show_dialog(
                    self, "Bad selected area: %s" % area_as_text, "Bad selected area"
                )
            elif not StringUtils.is_url_correct(url):
                show_dialog(self, "Bad url: %s" % url, "Bad url")
            else:
                dialog = DialogWithText(self, "Generating page object class...")
                handler = WxTextCtrlHandler(dialog.txt_ctrl)
                logger = Logger(log_to_console=False, handler=handler)

                dialog.Show()

                area = eval(area_as_text)
                generator = PageObjectGenerator(self.main_frame.get_browser(), logger)
                folder_path = self.main_frame.get_tmp_dir()

                def generate():
                    dialog.btn_ok.Disable()
                    po_class = generator.get_po_class_for_url(
                        url, class_name, folder_path, area
                    )
                    po_class.save(folder)
                    logger.info("Saving class '%s'..." % po_class.name)
                    logger.info("Saved file: %s" % po_class.file_path)
                    logger.info("Saved file: %s" % po_class.img_path)
                    logger.info("DONE")
                    dialog.btn_ok.Enable()

                thread = Thread(target=generate)
                thread.setDaemon(True)
                thread.start()



import os
import traceback

from easelenium.ui.editor.field_context_menu import FieldContextMenu
from easelenium.ui.editor.utils import FieldsTableAndTestFilesTabs, PyFileUI, TestFileUI
from easelenium.ui.file_utils import is_correct_python_file, read_file
from easelenium.ui.generator.page_object_class import PageObjectClass
from easelenium.ui.parser.parsed_class import (
    ParsedBrowserClass,
    ParsedMouseClass,
    ParsedPageObjectClass,
)
from easelenium.ui.root_folder import RootFolder
from easelenium.ui.utils import FLAG_ALL_AND_EXPAND
from easelenium.ui.widgets.image.image_with_elements import ImageWithElements
from easelenium.ui.widgets.utils import (
    ImageAndTableHelper,
    show_dialog,
    show_dialog_path_doesnt_exist,
    show_error_dialog,
)
from wx import (
    ALL,
    CB_READONLY,
    EVT_BUTTON,
    EVT_COMBOBOX,
    EVT_MOTION,
    EVT_RIGHT_DOWN,
    HORIZONTAL,
    ID_OK,
    SP_3D,
    SP_LIVE_UPDATE,
    BoxSizer,
    Button,
    ComboBox,
    FileDialog,
    GridBagSizer,
    Panel,
    SplitterWindow,
    StaticText,
)


class EditorTab(Panel):
    def __init__(self, parent):
        Panel.__init__(self, parent)
        sizer = GridBagSizer(5, 5)
        self.SetSizer(sizer)

        self.__cur_po_class = None
        self.__create_widgets()

    def __create_widgets(self):
        sizer = self.GetSizer()

        # Next row
        inner_sizer = BoxSizer(HORIZONTAL)
        label = StaticText(self, label="Class path:")
        inner_sizer.Add(label, flag=ALL)

        self.cb_class_path = ComboBox(self, style=CB_READONLY)
        self.cb_class_path.Bind(EVT_COMBOBOX, self.__on_load_po_class)
        inner_sizer.Add(self.cb_class_path, 1, flag=FLAG_ALL_AND_EXPAND)

        self.btn_reload = Button(self, label="Reload")
        self.btn_reload.Bind(EVT_BUTTON, self.__on_load_po_class)
        inner_sizer.Add(self.btn_reload, flag=ALL)

        self.btn_open_class = Button(self, label="Open class")
        self.btn_open_class.Bind(EVT_BUTTON, self.__open_class)
        inner_sizer.Add(self.btn_open_class, flag=ALL)

        row = 0
        sizer.Add(inner_sizer, pos=(row, 0), flag=FLAG_ALL_AND_EXPAND)

        # Next row
        row += 1
        splitter = SplitterWindow(self, style=SP_3D | SP_LIVE_UPDATE)

        self.image_panel = ImageWithElements(splitter)
        self.image_panel.static_bitmap.Bind(EVT_MOTION, self.__on_mouse_move)
        self.image_panel.static_bitmap.Bind(EVT_RIGHT_DOWN, self.__on_right_click)

        self.table_and_test_file_tabs = FieldsTableAndTestFilesTabs(splitter, self)

        splitter.SplitHorizontally(self.image_panel, self.table_and_test_file_tabs)
        sizer.Add(splitter, pos=(row, 0), flag=FLAG_ALL_AND_EXPAND)

        sizer.AddGrowableRow(row, 1)
        sizer.AddGrowableCol(0, 1)

    def __get_parsed_classes(self, field):
        classes = ParsedPageObjectClass.get_parsed_classes(
            self.__cur_po_class.file_path
        )
        if len(classes) > 0 and len(classes[0].methods) == 0:
            classes = []
        if field:
            classes += ParsedMouseClass.get_parsed_classes()
            classes += ParsedBrowserClass.get_parsed_classes()
        return classes

    def show_content_menu(self, field):
        tabs = self.table_and_test_file_tabs.tabs
        count = tabs.GetPageCount()
        if count > 1:
            selected_tab = tabs.GetPage(tabs.GetSelection())
            if type(selected_tab) in (TestFileUI, PyFileUI):
                file_path = selected_tab.get_file_path()
                txt_ctrl_ui = tabs.GetPage(tabs.GetSelection())
                parsed_classes = self.__get_parsed_classes(field)
                context_menu = FieldContextMenu(
                    field, parsed_classes, file_path, txt_ctrl_ui
                )
                self.PopupMenu(context_menu)
                context_menu.Destroy()
            else:
                show_dialog(
                    self,
                    "Please select tab with test file.",
                    "Tab with test file was not selected",
                )
        else:
            show_dialog(
                self,
                "Please create/open test file.",
                "Test file was not created/opened",
            )

    def __on_load_po_class(self, evt):
        path = self.cb_class_path.GetValue()
        if len(path) > 0:
            self.__load_po_class(path)

    def __on_right_click(self, evt):
        field = self.__get_current_field(evt)
        self.show_content_menu(field)

    def __on_mouse_move(self, evt):
        if self.__cur_po_class:
            ImageAndTableHelper.select_field_on_mouse_move(
                evt,
                self.__cur_po_class.fields,
                self.image_panel,
                self.table_and_test_file_tabs.table,
            )

    def __get_current_field(self, evt):
        return self.image_panel.get_field(evt.GetPosition())

    def __open_class(self, evt):
        folder = self.GetTopLevelParent().get_root_folder()
        if folder:
            if RootFolder.PO_FOLDER in os.listdir(folder):
                folder = os.path.join(folder, RootFolder.PO_FOLDER)
            dialog = FileDialog(self, defaultDir=folder, wildcard="*.py")
            if dialog.ShowModal() == ID_OK:
                self.__load_po_class(dialog.GetPath())
        else:
            show_dialog_path_doesnt_exist(self, folder)

    def __load_po_class(self, path):
        self.table_and_test_file_tabs.table.clear_table()

        if not os.path.exists(path):
            show_dialog_path_doesnt_exist(self, path)
        if not is_correct_python_file(path):
            show_dialog(self, "File name is incorrect: %s" % path, "Bad file name")
        else:
            folder = os.path.dirname(path)
            files = [
                os.path.join(folder, p)
                for p in os.listdir(folder)
                if is_correct_python_file(p)
            ]
            self.cb_class_path.Clear()
            self.cb_class_path.AppendItems(files)
            self.cb_class_path.Select(files.index(path))
            try:
                self.__cur_po_class = PageObjectClass.parse_string_to_po_class(
                    read_file(path)
                )
                area = self.__cur_po_class.area
                self.image_panel.set_po_fields(self.__cur_po_class.fields)
                self.image_panel.load_image(self.__cur_po_class.img_path, area)

                self.cb_class_path.SetValue(self.__cur_po_class.file_path)

                self.table_and_test_file_tabs.load_po_class(self.__cur_po_class)
            except Exception:
                self.__cur_po_class = None
                show_error_dialog(
                    self, traceback.format_exc(), "Failed to open file %s" % path
                )



import os
import tempfile
import traceback
from threading import Thread

from easelenium.ui.file_utils import save_file
from easelenium.ui.parser.parsed_class import ParsedClass
from wx import ALL, EXPAND

FLAG_ALL_AND_EXPAND = ALL | EXPAND


def run_in_separate_thread(target, name=None, args=(), kwargs=None):
    thread = Thread(target=target, name=name, args=args, kwargs=kwargs)
    thread.start()
    return thread


def check_py_code_for_errors(code, *additional_python_paths):
    tmp_file = tempfile.mkstemp()
    save_file(tmp_file, code)
    formatted_exception = check_file_for_errors(tmp_file, *additional_python_paths)
    os.remove(tmp_file)
    return formatted_exception


def check_file_for_errors(path, *additional_python_paths):
    syspath = list(os.sys.path)

    for py_path in additional_python_paths:
        if os.path.exists(py_path):
            os.sys.path.append(py_path)

    try:
        ParsedClass.get_parsed_classes(path)
        os.sys.path = syspath
        return None
    except Exception:
        os.sys.path = syspath
        formatted_exc = traceback.format_exc()
        return formatted_exc



# coding=utf8
import os

from easelenium.ui.file_utils import safe_create_path, save_file
from easelenium.utils import LINESEP, get_match
from selenium.webdriver.common.by import By

# TODO: move to f string and get rid of u strings


def get_by_as_code_str(by):
    if by == By.LINK_TEXT:
        return "By.LINK_TEXT"
    elif by == By.CLASS_NAME:
        return "By.CLASS_NAME"
    elif by == By.CSS_SELECTOR:
        return "By.CSS_SELECTOR"
    elif by == By.XPATH:
        return "By.XPATH"
    elif by == By.ID:
        return "By.ID"
    else:
        raise NotImplementedError


def get_by_from_code_str(by_as_string):
    if by_as_string == "By.LINK_TEXT":
        return By.LINK_TEXT
    elif by_as_string == "By.CLASS_NAME":
        return By.CLASS_NAME
    elif by_as_string == "By.CSS_SELECTOR":
        return By.CSS_SELECTOR
    elif by_as_string == "By.XPATH":
        return By.XPATH
    elif by_as_string == "By.ID":
        return By.ID
    else:
        raise NotImplementedError


class PageObjectClassField(object):
    def __init__(self, name, by, selector, location, dimensions):
        self.name = name
        self.by = by
        self.selector = selector
        self.location = location
        self.dimensions = dimensions

    def __eq__(self, other):
        return (other is not None) and (
            self.name == other.name
            and self.by == other.by
            and self.selector == other.selector
        )

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"PageObjectClassField({self.__dict__})"


class PageObjectClass(object):
    IMAGE_FOLDER = "img"
    TEMPLATE = """# coding=utf8
from selenium.webdriver.common.by import By

from easelenium.base_page_object import BasePageObject


class {name}(BasePageObject):
    # Please do NOT remove auto-generated comments
    # Url: {url}
    # Area: {area}
    # File path: {file_path}
    # Image path: {img_path}
{fields_as_code}

"""

    def __init__(
        self,
        name,
        url,
        fields,
        area=None,
        file_path=None,
        img_path=None,
        img_as_png=None,
    ):
        self.name = name
        self.url = url
        self.fields = fields
        self.area = area
        self.file_path = file_path
        self.img_as_png = img_as_png
        self.img_path = img_path

    def save(self, new_folder=None):
        if new_folder:
            py_filename = os.path.basename(self.file_path)
            img_filename = os.path.basename(self.img_path)
            self.file_path = os.path.abspath(os.path.join(new_folder, py_filename))
            self.img_path = os.path.abspath(
                os.path.join(new_folder, self.IMAGE_FOLDER, img_filename)
            )
        safe_create_path(self.file_path)
        safe_create_path(self.img_path)
        save_file(self.file_path, self._get_file_content())
        save_file(self.img_path, self.img_as_png, False)

    def _get_file_content(self):
        kwargs = self.__dict__.copy()
        fields_as_code = self._get_fields_as_code()
        if len(fields_as_code.strip()) == 0:
            fields_as_code = "    pass" + LINESEP
        kwargs["fields_as_code"] = fields_as_code
        return self.TEMPLATE.format(**kwargs)

    def _get_fields_as_code(self):
        single_line = "    {name} = ({by_as_code}, u'{selector}') # {comment}"
        lines = []
        for field in self.fields:
            lines.append(
                single_line.format(
                    **{
                        "name": field.name,
                        "by_as_code": get_by_as_code_str(field.by),
                        "selector": field.selector.replace("'", "\\'"),
                        "comment": "location: %s dimensions: %s"
                        % (field.location, field.dimensions),
                    }
                )
            )

        return LINESEP.join(lines)

    @classmethod
    def parse_string_to_po_class(cls, string):
        # class {name}(object):
        # Please do NOT remove auto-generated comments
        # Url: {url}
        # Area: {area}
        # Image path: {img_path}
        name_regexp = r"class (\w+)\(BasePageObject\):"
        url_regexp = r"Url: (.+)"
        area_regexp = r"Area: \(?([\w, ]+)\)?"
        img_path_regexp = r"Image path: (.+)"
        file_path_regexp = r"File path: (.+)"
        fields_regexp = r"\s+(\w+) = (.+) # location: (.+) dimensions: (.+)"

        name = get_match(name_regexp, string)
        url = get_match(url_regexp, string)
        area = eval(get_match(area_regexp, string))
        img_path = get_match(img_path_regexp, string)
        file_path = get_match(file_path_regexp, string)
        tmp_fields = get_match(fields_regexp, string, False)
        fields = []

        if tmp_fields:
            for (
                field_name,
                field_by_and_selector,
                field_location,
                field_dimensions,
            ) in tmp_fields:
                by, selector = eval(field_by_and_selector)
                location = eval(field_location)
                dimensions = eval(field_dimensions)
                po_class_field = PageObjectClassField(
                    field_name, by, selector, location, dimensions
                )
                fields.append(po_class_field)

        return PageObjectClass(name, url, fields, area, file_path, img_path)

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.url == other.url
            and self.fields == other.fields
            and self.area == other.area
            and self.file_path == other.file_path
            and self.img_path == other.img_path
        )

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"PageObjectClass({self.__dict__})"



from wx import EVT_MOTION, Cursor, CURSOR_HAND, CURSOR_ARROW, Point, Rect

from easelenium.ui.widgets.image.selectable_image import SelectableImagePanel


class ImageWithElements(SelectableImagePanel):
    def __init__(self, parent):
        SelectableImagePanel.__init__(self, parent)
        self.static_bitmap.Bind(EVT_MOTION, self.on_mouse_move)
        self.__po_fields = None
        self.selected_field = None

    def on_mouse_move(self, evt):
        field = self.get_field(evt.GetPosition())
        if self.selected_field != field:
            self.draw_selected_field(field)
            self.selected_field = field

    def set_po_fields(self, fields):
        self.__po_fields = fields

    def draw_selected_field(self, field, focus=False):
        end_pos = None
        start_pos = None
        if field:
            end_pos = (
                field.location[0] + field.dimensions[0],
                field.location[1] + field.dimensions[1],
            )
            start_pos = field.location
            cursor = Cursor(CURSOR_HAND)
        else:
            start_pos = (0, 0)
            if self.original_bitmap:
                end_pos = (
                    self.original_bitmap.GetWidth(),
                    self.original_bitmap.GetHeight(),
                )
            cursor = Cursor(CURSOR_ARROW)
        if start_pos and end_pos:
            self._draw_selected_area(start_pos, end_pos)
        self.static_bitmap.SetCursor(cursor)

        if field and focus:
            size = self.GetClientSize()
            x, y = field.location
            w, h = field.dimensions
            scroll_x = (x + w / 2 - size.GetWidth() / 2) / self.MIN_SCROLL
            scroll_y = (y + h / 2 - size.GetHeight() / 2) / self.MIN_SCROLL
            self.Scroll(scroll_x, scroll_y)

    def get_field(self, position):
        position = self._get_fixed_position(position)
        position = Point(*position)
        if self.__po_fields:
            fields_sorted_by_dimensions = sorted(
                self.__po_fields, key=lambda f: f.dimensions
            )
            for field in fields_sorted_by_dimensions:
                field_x, field_y = field.location
                w, h = field.dimensions

                if Rect(field_x, field_y, w, h).Contains(position):
                    return field
            return None
        else:
            return None



from wx import (
    ScrolledWindow,
    BoxSizer,
    VERTICAL,
    StaticBitmap,
    Image,
    BITMAP_TYPE_ANY,
    Bitmap,
    Rect,
    MemoryDC,
    BufferedDC,
    BLACK_PEN,
    TRANSPARENT_BRUSH,
    GREY_PEN,
    NullBitmap,
)

from easelenium.ui.utils import FLAG_ALL_AND_EXPAND


class ImagePanel(ScrolledWindow):
    MIN_SCROLL = 10

    def __init__(self, parent):
        ScrolledWindow.__init__(self, parent)

        self.wx_image = None
        self.original_bitmap = None
        self.greyscaled_bitmap = None
        self.img_path = None

        sizer = BoxSizer(VERTICAL)
        self.static_bitmap = StaticBitmap(self)
        sizer.Add(self.static_bitmap, 1, flag=FLAG_ALL_AND_EXPAND)

        self.SetSizer(sizer)

    def was_image_loaded(self):
        return (
            self.img_path
            and self.wx_image
            and self.original_bitmap
            and self.greyscaled_bitmap
        )

    def get_image_dimensions(self):
        return (self.original_bitmap.GetWidth(), self.original_bitmap.GetHeight())

    def load_image(self, path, area=None):
        self.Scroll(0, 0)
        self.img_path = path
        self.wx_image = Image(path, BITMAP_TYPE_ANY)
        width = self.wx_image.GetWidth()
        height = self.wx_image.GetHeight()
        if area:
            x, y, w, h = area
            bitmap = Bitmap(self.wx_image)
            bitmap_to_draw = bitmap.GetSubBitmap(Rect(x, y, w, h))

            bitmap = (
                bitmap.ConvertToImage()
                .ConvertToGreyscale(0.156, 0.308, 0.060)
                .ConvertToBitmap()
            )

            self.original_bitmap = self._get_bitmap(
                bitmap, bitmap_to_draw, x, y, w, h, False
            )
        else:
            self.original_bitmap = Bitmap(self.wx_image)
        self.greyscaled_bitmap = (
            self.original_bitmap.ConvertToImage()
            .ConvertToGreyscale(0.209, 0.411, 0.080)
            .ConvertToBitmap()
        )

        self.static_bitmap.SetBitmap(self.original_bitmap)
        self.SetScrollbars(
            self.MIN_SCROLL,
            self.MIN_SCROLL,
            width / self.MIN_SCROLL,
            height / self.MIN_SCROLL,
        )

    def _get_bitmap(self, bitmap, bitmap_to_draw, x, y, w, h, draw_frame=True):
        bitmap = bitmap.GetSubBitmap(Rect(0, 0, bitmap.GetWidth(), bitmap.GetHeight()))

        dc = MemoryDC()
        bdc = BufferedDC(dc)
        bdc.SelectObject(bitmap)
        bdc.DrawBitmap(bitmap_to_draw, x, y)

        if draw_frame:
            # Black rect to support white pages
            bdc.SetPen(BLACK_PEN)
            bdc.SetBrush(TRANSPARENT_BRUSH)
            bdc.DrawRectangle(x, y, w, h)

            # Grey rect to support black pages
            bdc.SetPen(GREY_PEN)
            bdc.SetBrush(TRANSPARENT_BRUSH)
            bdc.DrawRectangle(x + 1, y + 1, w - 2, h - 2)

        bdc.SelectObject(NullBitmap)
        return bitmap



from threading import Thread

from easelenium.ui.generator.page_object_generator import PageObjectGenerator
from easelenium.ui.string_utils import StringUtils
from easelenium.ui.utils import FLAG_ALL_AND_EXPAND
from easelenium.ui.widgets.image.image_with_elements import ImageWithElements
from easelenium.ui.widgets.table import Table
from easelenium.ui.widgets.utils import (
    DialogWithText,
    ImageAndTableHelper,
    WxTextCtrlHandler,
    show_dialog,
)
from easelenium.utils import Logger
from wx import (
    EVT_BUTTON,
    EVT_MOTION,
    SP_3D,
    SP_LIVE_UPDATE,
    Button,
    GridBagSizer,
    Panel,
    SplitterWindow,
)
from wx.grid import EVT_GRID_SELECT_CELL


class SelectorFinderTab(Panel):
    def __init__(self, parent):
        Panel.__init__(self, parent)

        self.main_frame = self.GetTopLevelParent()
        self.po_fields = None

        self.__create_widgets()

    def __create_widgets(self):
        sizer = GridBagSizer(5, 5)

        row = 0
        col = 1
        self.bth_reload_img = Button(self, label="Reload image")
        self.bth_reload_img.Bind(EVT_BUTTON, self.__load_img)
        sizer.Add(self.bth_reload_img, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.bth_reload_selectors = Button(self, label="Find selectors")
        self.bth_reload_selectors.Bind(EVT_BUTTON, self.__find_selectors)
        sizer.Add(self.bth_reload_selectors, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)
        # third row
        row += 1
        col = 0

        splitter = SplitterWindow(self, style=SP_3D | SP_LIVE_UPDATE)

        self.image_panel = ImageWithElements(splitter)
        self.image_panel.static_bitmap.Bind(EVT_MOTION, self.__on_mouse_move)

        self.table = Table(splitter)
        self.table.Bind(EVT_GRID_SELECT_CELL, self.__on_cell_select)

        splitter.SplitHorizontally(self.image_panel, self.table)
        sizer.Add(splitter, pos=(row, col), span=(1, 3), flag=FLAG_ALL_AND_EXPAND)

        sizer.AddGrowableCol(0, 1)
        sizer.AddGrowableRow(row, 1)

        self.SetSizer(sizer)

    def __update_table(self):
        if self.po_fields:
            self.table.load_data(self.po_fields)

    def __on_mouse_move(self, evt):
        ImageAndTableHelper.select_field_on_mouse_move(
            evt, self.po_fields, self.image_panel, self.table
        )

    def __on_cell_select(self, evt):
        self.table.selected_row = evt.GetRow()
        self.image_panel.draw_selected_field(self.table.get_selected_data(), True)
        evt.Skip()

    def __load_img(self, evt=None):
        browser = self.main_frame.get_browser()
        if browser:
            img_path = browser.save_screenshot(self.main_frame.get_tmp_dir())
            self.image_panel.load_image(img_path)
            self.main_frame.set_url(browser.get_current_url())

    def __find_selectors(self, evt):
        browser = self.main_frame.get_browser()
        if browser:
            url = self.main_frame.get_url()
            if not StringUtils.is_url_correct(url):
                show_dialog(self, "Bad url: %s" % url, "Bad url")
            else:
                dialog = DialogWithText(self, "Finding selectors...")
                handler = WxTextCtrlHandler(dialog.txt_ctrl)
                logger = Logger(log_to_console=False, handler=handler)

                dialog.Show()

                generator = PageObjectGenerator(browser, logger)

                def find_selectors():
                    dialog.btn_ok.Disable()
                    self.po_fields = generator.get_all_po_fields(url, None)
                    logger.info("DONE")
                    self.__update_table()
                    dialog.btn_ok.Enable()

                thread = Thread(target=find_selectors)
                thread.setDaemon(True)
                thread.start()



import codecs
import os

__ENCODING = "utf8"
__WRITE_MODE = "wb"
__READ_MODE = "rb"


def is_correct_python_file(filename):
    name = os.path.basename(filename)
    return name.endswith(".py") and not name.startswith("__")


def check_if_path_exists(path):
    if not os.path.exists(path):
        raise Exception("Path not found '%s'" % path)
    return True


def safe_remove_path(path):
    if os.path.exists(path):
        os.remove(path)
        return True
    return False


def safe_create_path(path):
    if os.path.exists(path):
        return False
    is_dir = len(os.path.splitext(path)[-1]) == 0
    if is_dir:
        os.makedirs(path)
    else:
        basedir = os.path.dirname(path)
        if not os.path.exists(basedir):
            os.makedirs(basedir)
        codecs.open(path, __WRITE_MODE, __ENCODING).close()


def save_file(path, content, is_text=True):
    if is_text:
        with codecs.open(path, __WRITE_MODE, encoding=__ENCODING) as f:
            f.write(content)
    else:
        with open(path, __WRITE_MODE) as f:
            f.write(content)


def read_file(path):
    with codecs.open(path, __READ_MODE, encoding=__ENCODING) as f:
        return f.read()


def get_list_of_files(path, recursively=False):
    files = []
    for f in os.listdir(path):
        file_path = os.path.join(path, f)
        if os.path.isdir(file_path) and recursively:
            files += get_list_of_files(file_path, recursively)
        else:
            files.append(file_path)
    return files



import os

from easelenium.ui.file_utils import save_file


class RootFolder(object):
    PO_FOLDER = "page_objects"
    TESTS_FOLDER = "tests"
    REPORTS = "reports"
    INIT_PY = "__init__.py"

    @classmethod
    def prepare_folder(cls, path):
        if os.path.isdir(path):
            files_and_folders = os.listdir(path)

            for folder in [cls.PO_FOLDER, cls.TESTS_FOLDER, cls.REPORTS]:
                if folder not in files_and_folders:
                    os.mkdir(os.path.join(path, folder))

            save_file(os.path.join(path, cls.PO_FOLDER, cls.INIT_PY), "")
        else:
            raise Exception("'%s' is not a folder." % path)



import logging
import os
import re
import sys
from datetime import datetime
from random import choice

from loguru import logger

LINESEP = os.linesep


def get_match(regexp, string, single_match=True):
    found = re.findall(regexp, string)
    if len(found) > 0:
        return found[0] if single_match else found
    else:
        return None


def is_windows():
    return sys.platform.startswith("win")


def get_timestamp():
    timetuple = datetime.now().timetuple()
    timestamp = "%d%02d%02d%02d%02d%02d" % timetuple[:6]
    return timestamp


def is_string(obj):
    return type(obj) == str


def get_random_value(_list, *val_to_skip):
    _tmp = list(_list)
    for skipped in val_to_skip:
        _tmp.remove(skipped)
    value = choice(_tmp)
    return value


class Logger(object):
    def __init__(
        self,
        name=None,
        log_to_console=True,
        file_path=None,
        handler=None,
        level=logging.INFO,
    ):
        self.__logger = logger

        if log_to_console:
            self.__logger.add(sys.stdout, filter=name, level=level)

        if file_path:
            self.__logger.add(file_path, filter=name, level=level)

        if handler:
            self.__logger.add(handler, filter=name, level=level)

    def debug(self, msg, *args, **kwargs):
        self.__logger.info(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.__logger.info(msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        self.__logger.warn(msg, *args, **kwargs)


def get_class_name_from_file(path):
    filename, _ = os.path.splitext(os.path.basename(path))
    return "".join([w.capitalize() for w in filename.split("_")])


def get_py_file_name_from_class_name(class_name):
    words = re.findall("[A-Z]*[a-z0-9]*", class_name)
    words = [w for w in words if len(w) > 0]
    return "_".join(words).lower() + ".py"



from easelenium.ui.generator.page_object_class import get_by_as_code_str
from wx.grid import Grid


class Table(Grid):
    def __init__(self, parent):
        Grid.__init__(self, parent)

        self.selected_row = None
        self.__data = None
        self.__data_attrs = ["name", "by", "selector", "location", "dimensions"]

    def get_selected_data(self):
        return (
            self.__data[self.selected_row]
            if self.selected_row is not None
            else self.selected_row
        )

    def load_data(self, data):
        self.clear_table()

        self.__data = data
        count = len(self.__data)

        if not self.GetTable():
            self.CreateGrid(count, len(self.__data_attrs))
            table = self.GetTable()
        else:
            table = self.GetTable()
            table.AppendRows(count)
            table.AppendCols(len(self.__data_attrs))

        # filling headers
        for attr in self.__data_attrs:
            table.SetColLabelValue(self.__data_attrs.index(attr), attr.capitalize())

        # filling data
        for d in self.__data:
            j = self.__data.index(d)
            for attr in self.__data_attrs:
                value = getattr(d, attr)
                i = self.__data_attrs.index(attr)
                is_by = attr == self.__data_attrs[1]
                is_location_or_dimensions = type(value) in (tuple, list)
                if is_by:
                    value = get_by_as_code_str(value)
                elif is_location_or_dimensions:
                    value = str(value)

                table.SetValue(j, i, value)

        self.AutoSizeColumns()

    def clear_table(self):
        self.ClearGrid()

    def select_row(self, row):
        if row:
            self.selected_row = row
            self.SelectRow(row)
            self.Scroll(0, row)



# coding=utf8
import traceback
from unittest.case import TestCase

from easelenium.browser import Browser
from easelenium.utils import Logger, get_timestamp


class BaseTest(TestCase):
    TC_NAME_WIDTH = 100
    BROWSER_NAME = None
    FAILED_SCREENSHOT_FOLDER = None
    LOGGER = Logger(name="easyselenim.base_test.BaseTest")

    @classmethod
    def setUpClass(cls, **kwargs):
        super(BaseTest, cls).setUpClass()

        kwargs["browser_name"] = kwargs.get("browser_name") or cls.BROWSER_NAME
        kwargs["logger"] = kwargs.get("logger") or cls.LOGGER

        cls.logger = kwargs["logger"]
        cls.browser = Browser(**kwargs)

    @classmethod
    def tearDownClass(cls):
        super(BaseTest, cls).tearDownClass()
        cls.browser.quit()

    def setUp(self):
        TestCase.setUp(self)
        if self.browser.logger:
            name = self.id()
            symbols_before = "-" * int((self.TC_NAME_WIDTH - len(name) - 2) / 2)
            self.browser.logger.info(
                "{} {} {}".format(symbols_before, name, symbols_before)
            )

    def tearDown(self):
        failed = True
        if hasattr(self, "_outcome"):
            # python3
            failed = self._outcome and not self._outcome.success
        elif hasattr(self, "_resultForDoCleanups") and hasattr(
            self._resultForDoCleanups, "result"
        ):
            # nose
            failed = not self._resultForDoCleanups.result.wasSuccessful()

        if failed:
            name = self.id()
            filename = "_".join(
                [name, self.browser.get_browser_initials(), get_timestamp()]
            )
            try:
                self.browser.save_screenshot(
                    self.FAILED_SCREENSHOT_FOLDER, filename + ".png"
                )
            except Exception:
                formatted_exc = traceback.format_exc()
                self.browser.logger.info(formatted_exc)
        TestCase.tearDown(self)

        if self.browser.logger:
            self.browser.logger.info("-" * self.TC_NAME_WIDTH)



import inspect
import os
from importlib.machinery import SourceFileLoader
from pprint import pformat
from unittest.case import TestCase

from easelenium.browser import Browser, Mouse
from easelenium.utils import is_string


class ParsedClass(object):
    PROTECTED_PREFIX = "_"
    PRIVATE_PREFIX = "__"

    def __init__(self, name, class_obj, fields, methods):
        self.name = name
        self.class_obj = class_obj
        self.fields = fields
        self.methods = methods

    def get_value(self, name):
        return self.fields.get(name) or self.methods.get(name)

    def get_code(self, name_or_method):
        if inspect.ismethod(name_or_method):
            method = name_or_method
        else:
            method = self.get_value(name_or_method)
        return inspect.getsource(method)

    def get_arg_spec(self, name_or_method):
        if inspect.ismethod(name_or_method) or inspect.isfunction(name_or_method):
            method = name_or_method
        else:
            method = self.get_value(name_or_method)
        return inspect.getfullargspec(method)

    def get_module(self):
        return inspect.getmodule(self.class_obj)

    def get_source_file(self):
        return inspect.getsourcefile(self.class_obj)

    def __str__(self):
        return pformat(self.__dict__)

    def __repr__(self):
        return str(self)

    @classmethod
    def get_parsed_classes(cls, module_or_class_or_path):
        is_path = is_string(module_or_class_or_path)
        if inspect.ismodule(module_or_class_or_path) or is_path:
            if is_path:
                module_name = os.path.splitext(
                    os.path.basename(module_or_class_or_path)
                )[0]
                module_or_class_or_path = SourceFileLoader(
                    module_name, module_or_class_or_path
                ).load_module()

            cur_module = inspect.getmodule(module_or_class_or_path)
            classes = inspect.getmembers(
                module_or_class_or_path,
                lambda o: inspect.isclass(o) and inspect.getmodule(o) == cur_module,
            )
        elif inspect.isclass(module_or_class_or_path):
            classes = [(module_or_class_or_path.__name__, module_or_class_or_path)]
        else:
            raise NotImplementedError

        def filter_private_members(members):
            return [
                m
                for m in members
                if cls.PROTECTED_PREFIX not in m[0] or cls.PRIVATE_PREFIX not in m[0]
            ]

        parsed_classes = []
        for class_name, _class in classes:
            methods = inspect.getmembers(_class, inspect.ismethod)
            methods += inspect.getmembers(_class, inspect.isfunction)
            methods = filter_private_members(methods)
            fields = inspect.getmembers(_class, lambda o: not inspect.isroutine(o))
            fields = filter_private_members(fields)
            parsed_classes.append(
                ParsedClass(class_name, _class, dict(fields), dict(methods))
            )

        return parsed_classes


class ParsedBrowserClass(ParsedClass):
    __LOCATOR_NAME = "element"
    __GOOD_METHODS = (
        "switch_to_default_content",
        "get_title",
        "get_text",
        "get_attribute",
        "get_current_url",
    )
    # NOTE: methods which are wrapped(ex. 'stale_exception_wrapper')
    # should be added to __GOOD_METHODS
    # otherwise they won't be added to ParsedClass object

    @classmethod
    def get_parsed_classes(cls, module_or_class_or_path=None):
        parsed_classes = ParsedClass.get_parsed_classes(Browser)
        for _class in parsed_classes:
            _class.methods = dict(
                [
                    (n, v)
                    for n, v in _class.methods.items()
                    if cls.__LOCATOR_NAME in _class.get_arg_spec(n).args
                    or n in cls.__GOOD_METHODS
                ]
            )
        return parsed_classes


class ParsedMouseClass(ParsedClass):
    __LOCATOR_NAME = "element"

    @classmethod
    def get_parsed_classes(cls, module_or_class_or_path=None):
        parsed_classes = ParsedClass.get_parsed_classes(Mouse)
        for _class in parsed_classes:
            _class.methods = dict(
                [
                    (n, v)
                    for n, v in _class.methods.items()
                    if cls.__LOCATOR_NAME in _class.get_arg_spec(n).args
                ]
            )
        return parsed_classes


class ParsedPageObjectClass(ParsedClass):
    @classmethod
    def get_parsed_classes(cls, module_or_class_or_path):
        parsed_classes = ParsedClass.get_parsed_classes(module_or_class_or_path)
        super_class = ParsedClass.get_parsed_classes(TestCase)[0]

        def filter_class_data(class1, class2, methods_or_fields):
            return dict(
                [
                    (n, v)
                    for n, v in getattr(class1, methods_or_fields).items()
                    if n not in getattr(class2, methods_or_fields)
                ]
            )

        for _class in parsed_classes:
            _class.methods = filter_class_data(_class, super_class, "methods")
            _class.fields = filter_class_data(_class, super_class, "fields")

        return parsed_classes



import shutil
import traceback
from tempfile import mkdtemp

from easelenium.browser import Browser
from easelenium.ui.editor.editor_ui import EditorTab
from easelenium.ui.generator.generator_ui import GeneratorTab
from easelenium.ui.root_folder import RootFolder
from easelenium.ui.selector_finder.finder_ui import SelectorFinderTab
from easelenium.ui.string_utils import StringUtils
from easelenium.ui.test_runner_ui import TestRunnerTab
from easelenium.ui.utils import FLAG_ALL_AND_EXPAND
from easelenium.ui.widgets.utils import Tabs, show_dialog, show_error_dialog
from wx import (
    EVT_BUTTON,
    EVT_CLOSE,
    ID_OK,
    Button,
    Choice,
    DirDialog,
    Frame,
    GridBagSizer,
    Panel,
    StaticText,
    TextCtrl,
)


class MainFrame(Frame):
    def __init__(self, parent):
        Frame.__init__(self, parent)
        self.__browser = None
        self.__tmp_dir = mkdtemp()

        self.SetTitle("Easy Selenium UI")
        self.SetSize((800, 600))

        self.__create_widgets()
        self.Bind(EVT_CLOSE, self.__on_close)

    def __create_widgets(self):
        panel = Panel(self)

        sizer = GridBagSizer(5, 5)
        row = 0
        col = 0
        label = StaticText(panel, label="Root folder:")
        sizer.Add(label, pos=(row, col))

        col += 1
        col_width = 4
        self.__txt_root_path = TextCtrl(panel)
        sizer.Add(
            self.__txt_root_path,
            pos=(row, col),
            span=(1, col_width),
            flag=FLAG_ALL_AND_EXPAND,
        )

        col += col_width
        self.__btn_set_root = Button(panel, label="Set root folder")
        self.__btn_set_root.Bind(EVT_BUTTON, self.__set_root_folder)
        sizer.Add(self.__btn_set_root, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        row += 1
        col = 0
        label = StaticText(panel, label="Url:")
        sizer.Add(label, pos=(row, col))

        col += 1
        self.__txt_url = TextCtrl(
            panel, value="https://www.google.com/"
        )  # TODO: remove url
        sizer.Add(self.__txt_url, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        label = StaticText(panel, label="Browser:")
        sizer.Add(label, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.__cb_browser = Choice(panel, choices=Browser.get_supported_browsers())
        self.__cb_browser.Select(0)
        sizer.Add(self.__cb_browser, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.bth_open_url = Button(panel, label="Open url")
        self.bth_open_url.Bind(EVT_BUTTON, self.__open_url)
        sizer.Add(self.bth_open_url, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.bth_close_url = Button(panel, label="Close browser")
        self.bth_close_url.Bind(EVT_BUTTON, self.__close_browser)
        sizer.Add(self.bth_close_url, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        row += 1
        col = 0
        tabs = Tabs(
            panel,
            [
                (GeneratorTab, "Generator"),
                (EditorTab, "Editor"),
                (TestRunnerTab, "Test runner"),
                (SelectorFinderTab, "Selector finder"),
            ],
        )
        sizer.Add(tabs, pos=(row, col), span=(1, 6), flag=FLAG_ALL_AND_EXPAND)

        sizer.AddGrowableCol(1, 1)
        sizer.AddGrowableRow(2, 1)
        panel.SetSizer(sizer)
        self.Layout()

    def get_root_folder(self):
        text = self.__txt_root_path.GetValue()
        if text and len(text) > 0:
            return text
        else:
            return None

    def __set_root_folder(self, evt):
        dialog = DirDialog(self)
        if dialog.ShowModal() == ID_OK:
            path = dialog.GetPath()
            RootFolder.prepare_folder(path)
            self.__txt_root_path.SetValue(path)

    def __open_url(self, evt):
        url = self.__txt_url.GetValue()
        if StringUtils.is_url_correct(url):
            self.bth_open_url.Disable()

            name = self.get_browser_initials()

            try:
                if self.__browser and self.__browser.get_browser_initials() != name:
                    self.__browser.quit()
                    self.__browser = Browser(name)
                elif not self.__browser:
                    self.__browser = Browser(name)
            except Exception:
                show_error_dialog(
                    self, traceback.format_exc(), "Failed to open browser"
                )
                self.__browser = None

            if self.__browser:
                self.__browser.open(url)
            # TODO: if generator or selector -> load image
            self.bth_open_url.Enable()
        else:
            show_dialog(self, "Bad url: %s" % url, "Bad url")

    def __close_browser(self, evt):
        if self.__browser:
            self.__browser.quit()
            self.__browser = None

    def __on_close(self, evt):
        self.__close_browser(evt)

        shutil.rmtree(self.__tmp_dir, ignore_errors=True)
        self.Destroy()

    def get_url(self):
        return self.__txt_url.GetValue()

    def set_url(self, url):
        self.__txt_url.SetValue(url)

    def get_browser(self):
        return self.__browser

    def get_browser_initials(self):
        return self.__cb_browser.GetStringSelection()

    def get_tmp_dir(self):
        return self.__tmp_dir



#!/usr/bin/env python3

import os
import sys
from pytest import main

sys.path.append(os.path.join(os.path.dirname(__file__), "..", ".."))

from easelenium.browser import Browser


class easeleniumPlugin(object):
    def pytest_addoption(self, parser):
        group = parser.getgroup("easelenium")
        group.addoption(
            "--browser",
            dest="BROWSER",
            help="Specify browser by using initials. If value was not passed then 'ff' will be used. ",
            choices=Browser.get_supported_browsers(),
        )

    def pytest_configure(self, config):
        Browser.DEFAULT_BROWSER = config.option.BROWSER


if __name__ == "__main__":
    sys.exit(main(plugins=[easeleniumPlugin()]))



import logging
from threading import Event
from time import sleep

from easelenium.ui.utils import FLAG_ALL_AND_EXPAND, run_in_separate_thread
from easelenium.ui.widgets.image.image_with_elements import ImageWithElements
from easelenium.utils import LINESEP
from wx import (
    CAPTION,
    CENTER,
    DEFAULT_DIALOG_STYLE,
    EVT_BUTTON,
    GA_HORIZONTAL,
    GA_SMOOTH,
    HSCROLL,
    ID_OK,
    OK,
    RESIZE_BORDER,
    STAY_ON_TOP,
    TE_MULTILINE,
    TE_READONLY,
    VERTICAL,
    BoxSizer,
    Button,
    CallAfter,
    Dialog,
    Gauge,
    MessageDialog,
    Notebook,
    StaticText,
    TextCtrl,
)


def show_error_dialog(parent, message, caption):
    return DialogWithText(parent, caption, message).ShowModal()


def show_dialog(parent, message, caption, style=OK | CENTER):
    return MessageDialog(parent, message, caption, style).ShowModal()


def show_dialog_path_does_exist(parent, path):
    if path is None:
        path = ""
    msg = "Path already exists: '%s'" % path
    caption = "Bad path"
    return show_dialog(parent, msg, caption)


def show_dialog_path_doesnt_exist(parent, path):
    if path is None:
        path = ""
    msg = "Path doesn't exist: '%s'" % path
    caption = "Bad path"
    return show_dialog(parent, msg, caption)


def show_dialog_bad_name(parent, name, *expected_names):
    msg = "Bad name: '%s'" % name
    if len(expected_names) > 0:
        msg += LINESEP + "Expected names like:" + LINESEP + LINESEP.join(expected_names)
    caption = "Bad name"
    return show_dialog(parent, msg, caption)


class Tabs(Notebook):
    def __init__(self, parent, tabs_and_title=None):
        Notebook.__init__(self, parent)
        if tabs_and_title and len(tabs_and_title):
            for tab, title in tabs_and_title:
                self.AddPage(tab(self), title)

    def get_tabs_text(self, index=None):
        if index is None:
            index = self.GetSelection()
        text = self.GetPageText(index)
        return text

    def set_tabs_text(self, text, index=None):
        if index is None:
            index = self.GetSelection()
        self.SetPageText(index, text)


class WxTextCtrlHandler(logging.Handler):
    def __init__(self, ctrl):
        logging.Handler.__init__(self)
        self.ctrl = ctrl

    def emit(self, record):
        s = self.format(record) + LINESEP
        CallAfter(self.ctrl.AppendText, s)


class DialogWithText(Dialog):
    def __init__(self, parent, title, text=None):
        Dialog.__init__(
            self, parent, title=title, style=DEFAULT_DIALOG_STYLE | RESIZE_BORDER
        )
        self.SetTitle(title)
        self.SetSize(600, 400)

        sizer = BoxSizer(VERTICAL)

        self.txt_ctrl = TextCtrl(self, style=TE_MULTILINE | TE_READONLY | HSCROLL)
        if text:
            self.txt_ctrl.SetValue(text)
        sizer.Add(self.txt_ctrl, 1, flag=FLAG_ALL_AND_EXPAND)

        self.btn_ok = Button(self, label="OK")
        self.btn_ok.Bind(EVT_BUTTON, self.__close)
        sizer.Add(self.btn_ok, flag=CENTER)

        self.SetSizer(sizer)

    def __close(self, evt):
        CallAfter(self.EndModal, ID_OK)
        self.Hide()


class InfiniteProgressBarDialog(Dialog):
    def __init__(self, parent, title, text):
        Dialog.__init__(self, parent, title=title, style=CAPTION | STAY_ON_TOP)
        sizer = BoxSizer(VERTICAL)

        self.label = StaticText(self, label=text)
        sizer.Add(self.label, flag=FLAG_ALL_AND_EXPAND)

        self.gauge = Gauge(self, style=GA_SMOOTH | GA_HORIZONTAL)
        sizer.Add(self.gauge, flag=FLAG_ALL_AND_EXPAND)

        self.close_event = Event()

        def show_progress():
            while not self.close_event.is_set():
                sleep(0.05)
                self.gauge.Pulse()
            CallAfter(self.EndModal, ID_OK)

        self.SetSizerAndFit(sizer)

        run_in_separate_thread(show_progress)


class ImageAndTableHelper(object):
    @staticmethod
    def select_field_on_mouse_move(evt, po_fields, image_panel, table):
        field_before = image_panel.selected_field
        image_panel.set_po_fields(po_fields)
        ImageWithElements.on_mouse_move(image_panel, evt)
        field_after = image_panel.selected_field
        if field_after and field_after != field_before:
            row = po_fields.index(field_after)
            table.select_row(row)



class Mouse(object):
    def __init__(self, browser):
        self.browser = browser

    def left_click(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        self.left_click_by_offset(
            element,
            0,
            0,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def left_click_by_offset(
        self,
        element=None,
        xoffset=0,
        yoffset=0,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        actions = self.browser.get_action_chains()
        element = self.browser._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.browser.wait_for_visible(
            element=element,
        )

        if type(element) == tuple:
            element = self.browser.find_element(
                element=element,
            )

        self.browser._safe_log(
            "Click at '%s' by offset(%s,%s)", element, xoffset, yoffset
        )

        actions.move_to_element(element).move_by_offset(
            xoffset, yoffset
        ).click().perform()

    def hover(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        self.browser._safe_log("Hover at '%s'", element)

        self.hover_by_offset(
            element,
            0,
            0,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def hover_by_offset(
        self,
        element=None,
        xoffset=0,
        yoffset=0,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        actions = self.browser.get_action_chains()
        element = self.browser._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.browser.wait_for_visible(
            element=element,
        )

        element = self.browser.find_element(
            element=element,
        )

        self.browser._safe_log(
            "Mouse over '%s' by offset(%s,%s)", element, xoffset, yoffset
        )

        actions.move_to_element(element).move_by_offset(xoffset, yoffset).perform()

    def right_click(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        actions = self.browser.get_action_chains()
        element = self.browser._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.browser.wait_for_visible(
            element=element,
        )

        if type(element) == tuple:
            element = self.browser.find_element(
                element=element,
            )

        self.browser._safe_log(
            "Right click at '%s'",
            self.browser._to_string(
                element=element,
            ),
        )

        actions.context_click(element).perform()

    def right_click_by_offset(
        self,
        element=None,
        xoffset=0,
        yoffset=0,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        actions = self.browser.get_action_chains()
        element = self.browser._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.browser.wait_for_visible(
            element=element,
        )

        if type(element) == tuple:
            element = self.browser.find_element(
                element=element,
            )

        self.browser._safe_log(
            "Right click at '%s' by offset(%s,%s)", element, xoffset, yoffset
        )

        actions.move_to_element(element).move_by_offset(
            xoffset, yoffset
        ).context_click().perform()



#!/usr/bin/env python3

import os
import sys

from wx import App

sys.path.append(os.path.join(os.path.dirname(__file__), "..", ".."))
from easelenium.ui.main_ui import MainFrame


if __name__ == "__main__":
    app = App(False)
    main_ui = MainFrame(None)
    main_ui.Show()
    app.MainLoop()

# TODO: use relative path in generator
# TODO: better naming for generated fields
# TODO: find and remove duplicate code
# TODO: add wait_for_text_changed
# TODO: add wait_for_value_changed
# TODO: add wait_for_attribute_changed
# TODO: add parent to Browser methods
# TODO: changes are not applied when running test - if method is added to pagemodel and is used in test - NoMethod exception will be raised
# TODO: if test case updated - doesn't updated in test runner
# TODO: generator - when browser is closed - remove
# TODO: add support for not uniq selectors
# TODO: if selector is not uniq in 'Editor' use status text with notification: 'NOTE: this selector is not uniq' and show all possible object which could be found with this selector
# TODO: add enable/disable for uniq selector in 'Selector finder'
# TODO: add support for adding code where mouse is located
# TODO: when adding code auto wrap input if it is a string??
# TODO: move to Python3!
# TODO: add copy selector in 'Selector finder'
# TODO: if generation log window close - stop all generation
# TODO: no vertical scroll in Windows in Editor tab
# TODO: second import of page object is inserted incorrectly in Windows
# TODO: forbid setting root folder 'img' or 'page_objects'
# TODO: fix Windows GUI bugs:
#     if split is move to the top - not possible to resize it again
#     if click on reload image image is located in center , hover over element and image adjusted to the left
#     if test runner selected tree windows is not visible
# TODO: split browser class into smaller classes like Waits, Getters etc
# TODO: add support for headless drivers, mobile driver, remote execution
# TODO: add save_page_code to browser and add auto save on failed
# TODO: html option is not supported in Test runner in Windows



from easelenium.ui.context_menu import ContextMenu
from easelenium.ui.widgets.utils import show_dialog


class FieldContextMenu(ContextMenu):
    def __init__(self, field, parsed_classes, test_file, txt_ctrl_ui):
        self.__field = field
        self.__parsed_classes = parsed_classes
        self.__test_file = test_file
        self.__txt_ctrl_ui = txt_ctrl_ui
        data = self.__prepare_data_from_classes(parsed_classes)

        ContextMenu.__init__(self, data)
        self._bind_evt_menu(self.__on_menu_click)

    def __prepare_data_from_classes(self, parsed_classes):
        data = []
        for pc in parsed_classes:
            is_asserts = "assert_true" in pc.methods
            is_mouse = "hover" in pc.methods
            is_browser = "click" in pc.methods
            if is_asserts or is_mouse or not is_browser:
                if is_mouse:
                    name = "Mouse"
                else:
                    name = pc.name
                data.append((name, self.__prepare_context_data(pc.methods)))
                if len(parsed_classes) > 1:
                    data.append((ContextMenu.SEPARATOR_TEXT, None))
            else:
                browser_data = self.__prepare_context_data(pc.methods)
                for name, method in browser_data:
                    data.append((name, method))

        if not parsed_classes:
            data.append(("assert", None))
        return data

    def __prepare_context_data(self, initial_data):
        if type(initial_data) == dict:
            initial_data = initial_data.items()
        # needs to be initially sorted so that submenu items will be sorted as well
        initial_data = sorted(initial_data, key=lambda x: x[0])

        data = {}

        def append_to_data(submenu_text, item_text, func):
            if submenu_text in data:
                data[submenu_text] += [(item_text, func)]
            else:
                data[submenu_text] = [(item_text, func)]

        for text, func in initial_data:
            if text.startswith("_") or text.startswith("find"):
                continue

            if "dropdown" in text:
                append_to_data("dropdowns", text, func)
            elif text.startswith("get"):
                append_to_data("getters", text, func)
            elif text.startswith("wait"):
                append_to_data("waits", text, func)
            else:
                data[text] = func

        data = sorted(data.items(), key=lambda x: x[0])
        return data

    def __on_menu_click(self, evt):
        if self.__test_file:
            if self.__txt_ctrl_ui.has_one_or_more_methods_or_test_cases():
                item_data = self._get_menu_item_data(evt.GetId())
                method_name = item_data.text
                method = item_data.func
                for pc in self.__parsed_classes:
                    if method in pc.methods.values():
                        arg_spec = pc.get_arg_spec(method)
                        self.__txt_ctrl_ui.append_method_call(
                            field=self.__field,
                            method_name=method_name,
                            method=method,
                            arg_spec=arg_spec,
                        )
                        break
                if method_name == "assert":
                    self.__txt_ctrl_ui.append_method_call(method_name=method_name)
            else:
                show_dialog(
                    self.GetParent(),
                    "Please create method or test case.",
                    "Class doesn't have any methods or test cases",
                )
        else:
            show_dialog(
                self.GetParent(),
                "Please select or create test file.",
                "Test file was not selected",
            )



import os
import re


class RegExps(object):
    CLASS_NAME = r"^[A-Z][a-zA-Z0-9]*$"
    TEST_FILE_NAME = r"^[a-z_\d]+_test.py$"
    TEST_CASE_NAME = r"^test_[a-z_\d]+$"
    METHOD_NAME = r"^[a-z][a-z_\d]+$"
    AREA = r"\( *\d+ *, *\d+ *, *\d+ *, *\d+ *\)"
    URL = r"https?://.+"


class StringUtils(object):
    @classmethod
    def is_test_file_name_correct(cls, file_name):
        file_name = os.path.basename(file_name)
        return cls.does_text_match_regexp(file_name, RegExps.TEST_FILE_NAME)

    @classmethod
    def is_area_correct(cls, area_as_string):
        return cls.does_text_match_regexp(area_as_string, RegExps.AREA)

    @classmethod
    def is_test_case_name_correct(cls, test_case_name):
        return cls.does_text_match_regexp(test_case_name, RegExps.TEST_CASE_NAME)

    @classmethod
    def is_method_name_correct(cls, test_case_name):
        return cls.does_text_match_regexp(test_case_name, RegExps.METHOD_NAME)

    @classmethod
    def is_class_name_correct(cls, class_name):
        return cls.does_text_match_regexp(class_name, RegExps.CLASS_NAME)

    @classmethod
    def is_url_correct(cls, url):
        return cls.does_text_match_regexp(url, RegExps.URL)

    @classmethod
    def does_text_match_regexp(cls, text, regexp):
        return re.match(regexp, text) is not None



# coding=utf8
import os
import re

from easelenium.ui.file_utils import check_if_path_exists, read_file
from easelenium.ui.generator.page_object_class import (
    PageObjectClass,
    PageObjectClassField,
)
from easelenium.ui.root_folder import RootFolder
from easelenium.utils import get_py_file_name_from_class_name
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from wx import Point, Rect

# TODO: when generating link_text - escape new lines


class PageObjectGenerator(object):
    GET_XPATH_USING_JS = read_file(
        os.path.join(os.path.dirname(__file__), "get_xpath.js")
    )
    ELEMENTS_SELECTOR = (
        By.CSS_SELECTOR,
        "[onclick], [jsaction], a, select, button, input, "
        "span, p, h1, h2, h3, h4, h5, h6, "
        "frame, iframe, .btn",
    )
    FRAMES_SELECTOR = (By.CSS_SELECTOR, "frame, iframe")

    def __init__(self, browser, logger=None):
        self.browser = browser
        self.logger = logger

    def __log(self, *msgs):
        if self.logger:
            self.logger.info(" ".join([str(m) for m in msgs]))

    def _get_name_for_field(self, element_or_by_and_selector):
        max_length = 30
        if isinstance(element_or_by_and_selector, WebElement):
            by_and_selector = self._get_selector(element_or_by_and_selector)
        else:
            by_and_selector = element_or_by_and_selector
        name = "_".join(
            [w.upper()[:max_length] for w in re.findall(r"\w+", by_and_selector[1])]
        )
        name = re.sub(r"_+", "_", name)
        if len(name) == 0:
            name = "BAD_NAME"
        return name

    def __is_correct_element(self, element, area, location_offset):
        bad_element_tags = ("option", "script")

        if area:
            if type(area) not in (tuple, list) or len(area) != 4:
                raise Exception("Bad area data '%s'" % str(area))
            area = Rect(*area)
            x, y = self.browser.get_location(element)
            if location_offset:
                # fixing location because it is located inside frame
                x += location_offset[0]
                y += location_offset[1]
            w, h = self.browser.get_dimensions(element)
            element_center = Point(int(x + w / 2), int(y + h / 2))
            is_element_inside = area.Contains(element_center)
        else:
            is_element_inside = True

        return (
            self.browser.is_visible(element)
            and element.tag_name not in bad_element_tags
            and is_element_inside
        )

    def __get_po_fields_from_page(self, area, location_offset=None):
        fields = []
        elements = self.browser.find_elements(self.ELEMENTS_SELECTOR)
        i = 1
        log_prefix = " " * 10
        for e in elements:
            if self.logger:
                self.__log(
                    "%5d/%d Trying to get PageObjectField for element %s"
                    % (i, len(elements), self.browser._to_string(e))
                )

            if self.__is_correct_element(e, area, location_offset):
                field = self.__get_pageobject_field(e, location_offset)
                if field:
                    self.__log(log_prefix, "PageObjectField:", field)
                    fields.append(field)
                else:
                    self.__log(log_prefix, "Failed to unique selector")
            else:
                self.__log(
                    log_prefix,
                    "Skipped - element is not supported/visible or is outside of area",
                )
            i += 1

        self.__log("Number of fields:", len(fields))
        return fields

    def get_all_po_fields(self, url, area=None):
        if self.browser.get_current_url() != url:
            self.browser.get(url)

        fields = []
        self.browser.switch_to_default_content()
        self.__log("Getting fields for main content")
        fields += self.__get_po_fields_from_page(area)

        for frame in self.browser.find_elements(self.FRAMES_SELECTOR):
            self.browser.switch_to_default_content()
            location_offset = self.browser.get_location(frame)

            if self.logger:
                self.__log("Getting fields for frame", self.browser._to_string(frame))

            self.browser.switch_to_frame(frame)

            fields += self.__get_po_fields_from_page(area, location_offset)

        self.__log("Total number of fields:", len(fields))

        names = []
        for field in fields:
            if field.name in names:
                uniq_name = False
                i = 0
                while not uniq_name:
                    new_name = "_".join([field.name, str(i)])
                    if new_name not in names:
                        uniq_name = True
                        field.name = new_name
                    i += 1
            names.append(field.name)

        return fields

    def get_po_class_for_url(self, url, class_name, folder_path, area=None):
        po_folder = os.path.join(folder_path, RootFolder.PO_FOLDER)
        img_folder = os.path.join(
            folder_path, RootFolder.PO_FOLDER, PageObjectClass.IMAGE_FOLDER
        )
        check_if_path_exists(folder_path)

        self.__log(
            "Generating PageObjectClass for url %s with area %s" % (url, str(area))
        )
        fields = self.get_all_po_fields(url, area)
        img_as_png = self.browser.get_screenshot_as_png()

        filename = get_py_file_name_from_class_name(class_name)
        file_path = os.path.join(po_folder, filename)
        img_path = os.path.join(img_folder, os.path.splitext(filename)[0] + ".png")

        return PageObjectClass(
            class_name, url, fields, area, file_path, img_path, img_as_png
        )

    def __get_pageobject_field(self, element, location_offset):
        by_and_selector = self._get_selector(element)
        if by_and_selector:
            by, selector = by_and_selector

            name = self._get_name_for_field(by_and_selector)
            is_frame = self.browser.find_element(element).tag_name in [
                "frame",
                "iframe",
            ]
            if is_frame:
                name = "FRAME_" + name
            name_starts_with_number = re.match(r"^\d+.+$", name)
            if name_starts_with_number:
                name = "N" + name

            location = self.browser.get_location(element)
            if location_offset:
                # fix location because it is inside frame
                location = (
                    location[0] + location_offset[0],
                    location[1] + location_offset[1],
                )
            dimensions = self.browser.get_dimensions(element)
            return PageObjectClassField(name, by, selector, location, dimensions)
        return None

    def _get_selector(self, element):
        for selector_func in (
            self._get_id_selector,
            self._get_link_text_selector,
            self._get_class_name_selector,
            self._get_css_selector,
            self._get_xpath_selector,
        ):
            by_and_selector = selector_func(element)
            if by_and_selector:
                return by_and_selector

        return None

    def _get_id_selector(self, element):
        _id = self.browser.get_id(element)
        if _id and len(self.browser.find_elements(by_id=_id)) == 1:
            return By.ID, _id
        else:
            return None

    def _get_css_selector(self, element):
        """Recursively tries to find unique CSS selector for given element.

        Goes up through DOM tree until HTML or BODY tag is found. If
        doesn't find unique selector returns None.

        """
        element = self.browser.find_element(element)
        cur_css_selector = ""
        _id = element.get_attribute("id").strip()

        if _id:
            cur_css_selector += "#%s" % _id
        else:
            class_name = element.get_attribute("class").strip()

            if class_name:
                class_name = re.sub(r"\s+", ".", class_name)
                cur_css_selector += ".%s" % class_name

        cur_el_tag = element.tag_name
        if cur_el_tag in ("body", "html") or len(cur_css_selector) == 0:
            return None
        else:
            by_and_css_selector = By.CSS_SELECTOR, cur_css_selector
            elements_count = self.browser.get_elements_count(by_and_css_selector)
            if elements_count == 1:
                return by_and_css_selector
            else:
                parent = self.browser.get_parent(element)
                css_parent_selector = self._get_css_selector(parent)
                if css_parent_selector:
                    new_css_selector = " > ".join(
                        (css_parent_selector[1], cur_css_selector)
                    )
                    return By.CSS_SELECTOR, new_css_selector
                else:
                    return None

    def _get_xpath_selector(self, element):
        return By.XPATH, self.browser.execute_js(self.GET_XPATH_USING_JS, element)

    def _get_link_text_selector(self, element):
        text = self.browser.get_text(element)
        if len(self.browser.find_elements((By.LINK_TEXT, text))) == 1 and len(text) > 1:
            return By.LINK_TEXT, text
        else:
            return None

    def _get_class_name_selector(self, element):
        class_name = self.browser.get_class(element)
        if (
            len(class_name) > 0
            and " " not in class_name
            and len(self.browser.find_elements((By.CLASS_NAME, class_name))) == 1
        ):
            return By.CLASS_NAME, class_name
        else:
            return None



from wx import EVT_LEFT_DOWN, EVT_MOTION, HORIZONTAL, VERTICAL, Rect

from easelenium.ui.widgets.image.image_panel import ImagePanel
from easelenium.utils import is_windows


class SelectableImagePanel(ImagePanel):
    def __init__(self, parent):
        ImagePanel.__init__(self, parent)
        self.static_bitmap.Bind(EVT_LEFT_DOWN, self.__on_mouse_down)
        self.static_bitmap.Bind(EVT_MOTION, self.on_mouse_move)

        self.__start_position = None
        self.__selected_area = None

    def __on_mouse_down(self, evt):
        self.__start_position = evt.GetPosition()
        if self.was_image_loaded():
            w = self.original_bitmap.GetWidth()
            h = self.original_bitmap.GetHeight()
            self._draw_selected_area((0, 0), (w, h))

    def on_mouse_move(self, evt):
        if self.was_image_loaded() and evt.Dragging() and self.__start_position:
            # TODO: doesn't work in Windows if window is scrolled
            start_position = self._get_fixed_position(self.__start_position)
            end_position = self._get_fixed_position(evt.GetPosition())
            self._draw_selected_area(start_position, end_position)

    def _get_fixed_position(self, position):
        fix_for_scrolling = 0 if is_windows() else self.MIN_SCROLL
        scroll_offset = (self.GetScrollPos(HORIZONTAL), self.GetScrollPos(VERTICAL))
        return (
            position[0] + scroll_offset[0] * fix_for_scrolling,
            position[1] + scroll_offset[1] * fix_for_scrolling,
        )

    def _draw_selected_area(self, start_pos, end_pos):
        # TODO: reimplement - remove lagging
        x = start_pos[0] if start_pos[0] < end_pos[0] else end_pos[0]
        y = start_pos[1] if start_pos[1] < end_pos[1] else end_pos[1]
        w = abs(end_pos[0] - start_pos[0])
        h = abs(end_pos[1] - start_pos[1])

        self.__selected_area = (x, y, w, h)
        selected_bitmap = self.original_bitmap.GetSubBitmap(Rect(x, y, w, h))
        bitmap = self._get_bitmap(self.greyscaled_bitmap, selected_bitmap, x, y, w, h)
        self.static_bitmap.SetBitmap(bitmap)

    def get_selected_area(self):
        if self.__selected_area:
            return self.__selected_area
        else:
            w, h = self.get_image_dimensions()
            return 0, 0, w, h



import os
import re

from easelenium.ui.file_utils import save_file
from easelenium.ui.parser.parsed_class import (
    ParsedBrowserClass,
    ParsedMouseClass,
    ParsedPageObjectClass,
)
from easelenium.ui.root_folder import RootFolder
from easelenium.ui.string_utils import StringUtils
from easelenium.ui.utils import (
    FLAG_ALL_AND_EXPAND,
    check_file_for_errors,
    check_py_code_for_errors,
)
from easelenium.ui.widgets.table import Table
from easelenium.ui.widgets.utils import (
    Tabs,
    show_dialog,
    show_dialog_bad_name,
    show_error_dialog,
)
from easelenium.utils import LINESEP, get_class_name_from_file, is_windows
from wx import (
    ALIGN_RIGHT,
    DEFAULT_DIALOG_STYLE,
    EVT_BUTTON,
    EVT_KEY_DOWN,
    FD_OPEN,
    FD_SAVE,
    FONTFAMILY_TELETYPE,
    HORIZONTAL,
    HSCROLL,
    ID_CANCEL,
    ID_OK,
    NORMAL,
    RESIZE_BORDER,
    TE_MULTILINE,
    TE_READONLY,
    VERTICAL,
    WXK_TAB,
    BoxSizer,
    Button,
    CallAfter,
    Dialog,
    FileDialog,
    Font,
    GridBagSizer,
    Panel,
    StaticText,
    TextCtrl,
    TextEntryDialog,
)
from wx.grid import EVT_GRID_CELL_RIGHT_CLICK, EVT_GRID_SELECT_CELL


class PyFileUI(Panel):
    CHANGED_PREFIX = "*"
    METHOD_TEMPLATE = """
    def {method_name}(self):
        pass

"""

    def __init__(self, parent, file_path, load_file=False):
        Panel.__init__(self, parent)
        self.grandparent = self.GetGrandParent()

        self.__file_path = file_path

        sizer = BoxSizer(VERTICAL)
        self.txt_test_file_path = TextCtrl(
            self, value=self.__file_path, style=TE_READONLY
        )
        sizer.Add(self.txt_test_file_path, 0, flag=FLAG_ALL_AND_EXPAND)

        self.txt_content = TextCtrl(self, style=TE_MULTILINE | HSCROLL)
        self.txt_content.Bind(EVT_KEY_DOWN, self.__on_text_change)
        if load_file:
            self.txt_content.LoadFile(self.__file_path)
        font_size = self.txt_content.GetFont().GetPointSize()
        self.txt_content.SetFont(Font(font_size, FONTFAMILY_TELETYPE, NORMAL, NORMAL))
        sizer.Add(self.txt_content, 1, flag=FLAG_ALL_AND_EXPAND)

        self.SetSizer(sizer)

    def get_file_path(self):
        return self.__file_path

    def __get_selected_tab_text(self):
        parent = self.GetParent()
        return parent.get_tabs_text()

    def __set_selected_tab_text(self, text):
        parent = self.GetParent()
        parent.set_tabs_text(text)

    def load_file(self, path):
        self.__file_path = path
        self.txt_test_file_path.SetValue(self.__file_path)
        self.txt_content.LoadFile(self.__file_path)

    def save_file(self):
        root_folder = self.GetTopLevelParent().get_root_folder()
        text = self.__get_selected_tab_text()
        if text.startswith(self.CHANGED_PREFIX):
            self.__set_selected_tab_text(text[1:])
            save_file(self.__file_path, self.txt_content.GetValue())
            formatted_exc = check_file_for_errors(self.__file_path, root_folder)
            if formatted_exc:
                show_error_dialog(self, formatted_exc, "File contains errors")

    def set_file_was_changed(self):
        text = self.__get_selected_tab_text()
        if not text.startswith(self.CHANGED_PREFIX):
            self.__set_selected_tab_text(self.CHANGED_PREFIX + text)

    def __on_text_change(self, evt):
        key = evt.GetUnicodeKey()
        ctrl_s_pressed = key == 83 and evt.ControlDown()
        if key == WXK_TAB:
            indent = "    "
            self.txt_content.WriteText(indent)
        elif ctrl_s_pressed:
            self.save_file()
        else:
            self.set_file_was_changed()
            evt.Skip()

    def append_text(self, text):
        content = self.txt_content.GetValue()
        pass_with_tab = "    pass"
        if pass_with_tab in content:
            content = content.replace(pass_with_tab, "").rstrip() + LINESEP
            self.txt_content.SetValue(content)

        self.txt_content.AppendText(text)
        self.set_file_was_changed()

    def insert_text(self, line, pos):
        self.txt_content.SetInsertionPoint(pos)
        self.txt_content.WriteText(line.rstrip() + LINESEP)
        self.txt_content.SetInsertionPointEnd()
        self.set_file_was_changed()

    def create_method(self, method_name):
        test_case = self.METHOD_TEMPLATE.format(method_name=method_name)
        self.append_text(test_case)

    def has_one_or_more_methods_or_test_cases(self):
        return len(re.findall(r"def [a-z_]+\(self.+:", self.txt_content.GetValue())) > 0

    def append_method_call(
        self, field=None, method_name=None, method=None, arg_spec=None
    ):
        # TODO: simplify
        assert field or method_name
        if method_name == "assert":
            var = "expression"
            dialog = MultipleTextEntry(self, "Please enter values", [var])
            if dialog.ShowModal() == ID_OK:
                self.append_text(f"\n        assert {dialog.values[var]}")
            return

        is_po_class_file_selected = type(self) == PyFileUI
        # removing arguments with default value
        args = arg_spec.args
        if arg_spec.defaults:
            args = args[: -len(arg_spec.defaults)]

        # removing 'self' argument
        self_txt = "self"
        if self_txt in args:
            args.remove(self_txt)

        po_class = self.grandparent.get_current_pageobject_class()
        lowered_class_name = po_class.name.lower()

        is_assert_method = method_name.startswith("assert")
        is_browser_method = (
            method_name in ParsedBrowserClass.get_parsed_classes()[0].methods
        )
        is_mouse_method = (
            method_name in ParsedMouseClass.get_parsed_classes()[0].methods
        )
        is_page_object_method = (
            method_name
            in ParsedPageObjectClass.get_parsed_classes(po_class.file_path)[0].methods
        )

        # replacing 'element' with correctly formatted string - self.obj.field
        element_txt = "element"
        if element_txt in args:
            element_index = args.index(element_txt)
            if is_po_class_file_selected:
                args[element_index] = "self.%s" % field.name
            else:
                args[element_index] = "self.%s.%s" % (lowered_class_name, field.name)

        if is_browser_method:
            caller = "self.browser"
        elif is_mouse_method:
            caller = "self.browser.mouse"
        elif is_page_object_method and not is_po_class_file_selected:
            caller = "self." + lowered_class_name
        elif is_assert_method:
            caller = None
        else:
            caller = "self"

        get_prefix = "get_"
        is_getter_method_and_no_args = method_name.startswith(get_prefix)
        method_kwargs = {}
        if is_getter_method_and_no_args:
            var = method_name.replace(get_prefix, "")
            method_call_template = (
                "        {var} = {caller}.{method}({method_args})" + LINESEP
            )
            method_kwargs["var"] = var
        elif caller is None:
            method_call_template = "        {method}({method_args})" + LINESEP
        else:
            method_call_template = "        {caller}.{method}({method_args})" + LINESEP

        if (
            len(args) > 1
            and (is_browser_method or is_mouse_method)
            or len(args) > 0
            and (is_assert_method or is_page_object_method)
        ):
            if is_assert_method or is_page_object_method:
                dialog = MultipleTextEntry(self, "Please enter values", args)
            else:
                dialog = MultipleTextEntry(self, "Please enter values", args[1:])
            if dialog.ShowModal() == ID_OK:
                for name, value in dialog.values.items():
                    args[args.index(name)] = value
                method_kwargs.update(
                    {
                        "caller": caller,
                        "method": method_name,
                        "method_args": ", ".join(args),
                    }
                )
                code_line = method_call_template.format(**method_kwargs)
                code = self.txt_content.GetValue() + LINESEP + code_line
                root_folder = self.GetTopLevelParent().get_root_folder()
                formatted_exception = check_py_code_for_errors(code, root_folder)

                if formatted_exception:
                    show_dialog(
                        self, formatted_exception, "Values are not Python expressions"
                    )
                else:
                    method_kwargs.update(
                        {
                            "caller": caller,
                            "method": method_name,
                            "method_args": ", ".join(args),
                        }
                    )
                    self.append_text(method_call_template.format(**method_kwargs))
        else:
            method_kwargs.update(
                {
                    "caller": caller,
                    "method": method_name,
                    "method_args": ", ".join(args),
                }
            )
            self.append_text(method_call_template.format(**method_kwargs))


class TestFileUI(PyFileUI):
    TEST_FILE_TEMPLATE = """# coding=utf8
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from easelenium.base_test import BaseTest


class {class_name}(BaseTest):

    @classmethod
    def setUpClass(cls):
        super({class_name}, cls).setUpClass()

    def setUp(self):
        self.browser.get(u'{url}')

"""

    def __init__(self, parent, test_file_path, po_class, load_file=False):
        PyFileUI.__init__(self, parent, test_file_path, load_file)

        self.__po_class = po_class

        if not load_file:
            initial_text = self.TEST_FILE_TEMPLATE.format(
                class_name=get_class_name_from_file(test_file_path), url=po_class.url
            )
            self.txt_content.SetValue(initial_text)

    def __fix_imports(self, po_class):
        root_folder = os.path.normpath(self.GetTopLevelParent().get_root_folder())
        path, _ = os.path.splitext(po_class.file_path)
        path = path.replace(root_folder, "").strip()
        paths = [p for p in os.path.normpath(path).split(os.sep) if len(p) > 0]
        correct_import = ".".join(paths)
        class_name = po_class.name
        import_line = "from {relative_import} import {class_name}".format(
            relative_import=correct_import, class_name=class_name
        )
        text = self.txt_content.GetValue()
        if import_line not in text:
            base_test_import = "import BaseTest"
            shift = 6 if is_windows() else len(LINESEP)  # windows hack
            pos = text.index(base_test_import) + len(base_test_import) + shift
            self.insert_text(import_line, pos)

    def __fix_class_initialization(self, po_class):
        class_name = po_class.name
        field_name = class_name.lower()
        field_line = (
            "        cls.{field_name} = {class_name}(cls.browser, cls.logger)".format(
                field_name=field_name, class_name=class_name
            )
        )

        text = self.txt_content.GetValue()
        if field_line not in text:
            class_def_end = "cls).setUpClass()"
            shift = 13 if is_windows() else len(LINESEP)  # windows hack
            pos = text.index(class_def_end) + len(class_def_end) + shift
            self.insert_text(field_line, pos)

    def append_method_call(
        self, field=None, method_name=None, method=None, arg_spec=None
    ):
        assert field or method_name
        if method_name != "assert":
            po_class = self.grandparent.get_current_pageobject_class()

            self.__fix_class_initialization(po_class)
            self.__fix_imports(po_class)

        PyFileUI.append_method_call(
            self, field=field, method_name=method_name, method=method, arg_spec=arg_spec
        )

    def create_new_test_case(self, test_case_name):
        self.create_method(test_case_name)


class FieldsTableAndTestFilesTabs(Panel):
    TAB_INDEX_FOR_TABLE = 0
    TAB_INDEX_FOR_PO_CLASS_FILE = 1

    def __init__(self, parent, editor_tab):
        Panel.__init__(self, parent)

        self.__editor_tab = editor_tab
        self.__cur_po_class = None

        sizer = GridBagSizer(5, 5)
        full_span = (1, 4)

        row = 0
        inner_sizer = BoxSizer(HORIZONTAL)
        self.btn_open_test_file = Button(self, label="Open test file")
        self.btn_open_test_file.Bind(EVT_BUTTON, self.__on_open_test_file)
        inner_sizer.Add(self.btn_open_test_file)

        self.btn_create_test_file = Button(self, label="Create test file")
        self.btn_create_test_file.Bind(EVT_BUTTON, self.__on_create_test_file)
        inner_sizer.Add(self.btn_create_test_file)

        self.btn_save_test_file = Button(self, label="Save current file")
        self.btn_save_test_file.Bind(EVT_BUTTON, self.__on_save_test_file)
        inner_sizer.Add(self.btn_save_test_file)

        inner_sizer.AddStretchSpacer(1)

        self.btn_create_test = Button(self, label="Create new method/test case")
        self.btn_create_test.Bind(EVT_BUTTON, self.__create_method_or_test)
        inner_sizer.Add(self.btn_create_test)
        sizer.Add(inner_sizer, pos=(row, 0), span=full_span, flag=FLAG_ALL_AND_EXPAND)

        row += 1
        self.tabs = Tabs(self, [(Table, "Fields' table")])
        self.table = self.tabs.GetPage(0)
        self.table.Bind(EVT_GRID_SELECT_CELL, self.__on_cell_click)
        self.table.Bind(EVT_GRID_CELL_RIGHT_CLICK, self.__on_cell_click)

        sizer.Add(self.tabs, pos=(row, 0), span=full_span, flag=FLAG_ALL_AND_EXPAND)

        sizer.AddGrowableCol(1, 1)
        sizer.AddGrowableRow(1, 1)
        self.SetSizer(sizer)

    def load_po_class(self, po_class):
        self.__cur_po_class = po_class
        self.__set_pageobject_class(self.__cur_po_class)

        file_name = os.path.basename(self.__cur_po_class.file_path)

        more_than_1_tab = self.tabs.GetPageCount() > 1
        if more_than_1_tab:
            py_file_ui = self.tabs.GetPage(self.TAB_INDEX_FOR_PO_CLASS_FILE)
            py_file_ui.load_file(self.__cur_po_class.file_path)
        else:
            py_file_ui = PyFileUI(self.tabs, self.__cur_po_class.file_path, True)
            self.tabs.AddPage(
                py_file_ui, os.path.basename(self.__cur_po_class.file_path)
            )

        if (
            not more_than_1_tab
            and self.tabs.GetSelection() != self.TAB_INDEX_FOR_PO_CLASS_FILE
        ):
            self.tabs.SetSelection(self.TAB_INDEX_FOR_PO_CLASS_FILE)

        self.tabs.set_tabs_text(file_name, self.TAB_INDEX_FOR_PO_CLASS_FILE)

    def get_current_po_class(self):
        return self.__cur_po_class

    def clear_table(self):
        self.table.ClearGrid()

    def get_current_pageobject_class(self):
        return self.__cur_po_class

    def __set_pageobject_class(self, po_class):
        self.__cur_po_class = po_class
        self.table.load_data(po_class.fields)

    def __create_method_or_test(self, evt):
        count = self.tabs.GetPageCount()
        if count > 1:
            page = self.tabs.GetPage(self.tabs.GetSelection())
            if type(page) == PyFileUI:
                self.__create_method(page)
            elif type(page) == TestFileUI:
                self.__create_test(page)
            else:
                show_dialog(
                    self,
                    "Selected tab is not supported"
                    + LINESEP
                    + "Please selected test file or page object class",
                    "Bad selected tab",
                )
        else:
            show_dialog(
                self,
                "Test file was not created." + LINESEP + "Please create a test file.",
                "Test file was not created",
            )

    def __create_method(self, page):
        modal = TextEntryDialog(self, "Enter method name", "Create method")
        if modal.ShowModal() == ID_OK:
            method_name = modal.GetValue()
            if StringUtils.is_method_name_correct(method_name):
                page.create_method(method_name)
            else:
                show_dialog_bad_name(self, method_name, "search", "login", "fill_data")

    def __create_test(self, page):
        modal = TextEntryDialog(self, "Enter test case name", "Create new test case")
        if modal.ShowModal() == ID_OK:
            test_case_name = modal.GetValue()
            if StringUtils.is_test_case_name_correct(test_case_name):
                page.create_new_test_case(test_case_name)
            else:
                show_dialog_bad_name(self, test_case_name, "test_search")

    def __on_save_test_file(self, evt):
        count = self.tabs.GetPageCount()
        if count > 1:
            page = self.tabs.GetPage(self.tabs.GetSelection())
            page.save_file()
        else:
            show_dialog(self, "Please create/open test file.", "Nothing to save")

    def __open_or_create_test_file(self, style):
        if self.__cur_po_class:
            folder = self.GetTopLevelParent().get_root_folder()
            if not folder:
                folder = os.path.dirname(self.__cur_po_class.file_path)
            elif RootFolder.TESTS_FOLDER in os.listdir(folder):
                folder = os.path.join(folder, RootFolder.TESTS_FOLDER)

            dialog = FileDialog(self, defaultDir=folder, style=style, wildcard="*.py")
            if dialog.ShowModal() == ID_OK:
                test_file = dialog.GetPath()
                load_file = style == FD_OPEN

                filename = os.path.basename(test_file)
                if StringUtils.is_test_file_name_correct(test_file):
                    test_file_ui = TestFileUI(
                        self.tabs, test_file, self.__cur_po_class, load_file
                    )
                    self.tabs.AddPage(test_file_ui, filename)
                    self.tabs.SetSelection(self.tabs.GetPageCount() - 1)

                    if style == FD_SAVE:
                        test_file_ui.set_file_was_changed()
                else:
                    show_dialog_bad_name(self, filename, "my_first_test.py")
        else:
            show_dialog(self, "Please select class file.", "Class file was not opened")

    def __on_open_test_file(self, evt):
        self.__open_or_create_test_file(FD_OPEN)

    def __on_create_test_file(self, evt):
        self.__open_or_create_test_file(FD_SAVE)

    def __on_cell_click(self, evt):
        self.table.selected_row = evt.GetRow()
        field = self.table.get_selected_data()
        if field:
            self.__editor_tab.image_panel.draw_selected_field(field, True)
            if evt.GetEventType() == EVT_GRID_CELL_RIGHT_CLICK.typeId and field:
                self.__editor_tab.show_content_menu()
        evt.Skip()


class MultipleTextEntry(Dialog):
    def __init__(self, parent, title, values):
        Dialog.__init__(
            self, parent, title=title, style=DEFAULT_DIALOG_STYLE | RESIZE_BORDER
        )
        self.values = None

        sizer = GridBagSizer(5, 5)
        row = 0
        self.labels = []
        self.txt_ctrls = []
        for value in values:
            label = StaticText(self, label=value)
            self.labels.append(label)
            sizer.Add(label, pos=(row, 0), flag=ALIGN_RIGHT)

            txtctrl = TextCtrl(self)
            self.txt_ctrls.append(txtctrl)
            sizer.Add(txtctrl, pos=(row, 1), flag=FLAG_ALL_AND_EXPAND)
            row += 1

        self.btn_cancel = Button(self, label="Cancel")
        self.btn_cancel.Bind(EVT_BUTTON, self.__on_btn)
        sizer.Add(self.btn_cancel, pos=(row, 0), flag=FLAG_ALL_AND_EXPAND)

        self.btn_ok = Button(self, label="OK")
        self.btn_ok.Bind(EVT_BUTTON, self.__on_btn)
        sizer.Add(self.btn_ok, pos=(row, 1), flag=FLAG_ALL_AND_EXPAND)

        sizer.AddGrowableCol(1)
        self.SetSizerAndFit(sizer)
        self.SetSize(400, self.GetSizeTuple()[1])

    def __on_btn(self, evt):
        errors = []
        obj = evt.GetEventObject()
        if obj == self.btn_ok:
            self.values = {}
            for txt_ctrl in self.txt_ctrls:
                label_ctrl = self.labels[self.txt_ctrls.index(txt_ctrl)]
                label = label_ctrl.GetLabel()
                value = txt_ctrl.GetValue()
                self.values[label] = value
                if len(value) == 0:
                    errors.append("Variable '%s' has empty value '%s'" % (label, value))

            return_code = ID_OK
        else:
            self.values = None
            return_code = ID_CANCEL

        if len(errors) > 0:
            show_dialog(self, LINESEP.join(errors), "Bad entered data")
        else:
            CallAfter(self.EndModal, ID_OK)
        return return_code



import os
import traceback
from subprocess import check_output

from easelenium.base_test import BaseTest
from easelenium.browser import Browser
from easelenium.ui.file_utils import get_list_of_files
from easelenium.ui.parser.parsed_class import ParsedClass
from easelenium.ui.root_folder import RootFolder
from easelenium.ui.utils import FLAG_ALL_AND_EXPAND, run_in_separate_thread
from easelenium.ui.widgets.utils import (
    DialogWithText,
    InfiniteProgressBarDialog,
    show_dialog_path_doesnt_exist,
    show_error_dialog,
)
from pytest import main
from wx import (
    DD_DIR_MUST_EXIST,
    EVT_BUTTON,
    EVT_CHECKBOX,
    FD_FILE_MUST_EXIST,
    FD_MULTIPLE,
    FD_OPEN,
    FD_OVERWRITE_PROMPT,
    FD_SAVE,
    FONTFAMILY_TELETYPE,
    HSCROLL,
    ID_OK,
    NORMAL,
    SP_3D,
    SP_LIVE_UPDATE,
    TE_MULTILINE,
    TE_READONLY,
    TR_HAS_BUTTONS,
    TR_SINGLE,
    VSCROLL,
    Button,
    CallAfter,
    CheckBox,
    Choice,
    DirDialog,
    FileDialog,
    Font,
    GridBagSizer,
    Panel,
    SplitterWindow,
    StaticText,
    TextCtrl,
)
from wx.lib.agw.customtreectrl import (
    EVT_TREE_ITEM_CHECKED,
    TR_AUTO_CHECK_CHILD,
    TR_AUTO_CHECK_PARENT,
    TR_AUTO_TOGGLE_CHILD,
    CustomTreeCtrl,
)


class RedirectText(object):
    def __init__(self, txt_ctrl):
        self.out = txt_ctrl

    def write(self, string):
        CallAfter(self.out.WriteText, string)

    def isatty(self):
        return False

    def flush(self):
        pass


class TestRunnerTab(Panel):
    def __init__(self, *args, **kwargs):
        Panel.__init__(self, *args, **kwargs)
        sizer = GridBagSizer(5, 5)

        row = 0
        col = 0
        self.cb_html_output = CheckBox(self, label="Report in HTML")
        self.cb_html_output.Bind(EVT_CHECKBOX, self.__on_check)
        sizer.Add(self.cb_html_output, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.txt_html_report = TextCtrl(self, style=TE_READONLY)
        self.txt_html_report.Disable()
        sizer.Add(
            self.txt_html_report, pos=(row, col), span=(1, 3), flag=FLAG_ALL_AND_EXPAND
        )

        col += 3
        self.btn_select_html = Button(self, label="Select HTML file")
        self.btn_select_html.Disable()
        self.btn_select_html.Bind(EVT_BUTTON, self.__on_select_file)
        sizer.Add(self.btn_select_html, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        row += 1
        col = 0
        self.cb_xml_output = CheckBox(self, label="Report in XML")
        self.cb_xml_output.Bind(EVT_CHECKBOX, self.__on_check)
        sizer.Add(self.cb_xml_output, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.txt_xml_report = TextCtrl(self, style=TE_READONLY)
        self.txt_xml_report.Disable()
        sizer.Add(
            self.txt_xml_report, pos=(row, col), span=(1, 3), flag=FLAG_ALL_AND_EXPAND
        )

        col += 3
        self.btn_select_xml = Button(self, label="Select XML file")
        self.btn_select_xml.Disable()
        self.btn_select_xml.Bind(EVT_BUTTON, self.__on_select_file)
        sizer.Add(self.btn_select_xml, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        row += 1
        col = 0
        self.cb_options = CheckBox(self, label="Additional options")
        self.cb_options.Bind(EVT_CHECKBOX, self.__on_check)
        sizer.Add(self.cb_options, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.txt_options = TextCtrl(self)
        self.txt_options.Disable()
        sizer.Add(
            self.txt_options, pos=(row, col), span=(1, 3), flag=FLAG_ALL_AND_EXPAND
        )

        col += 3
        self.btn_nose_help = Button(self, label="Show help")
        self.btn_nose_help.Bind(EVT_BUTTON, self.__on_show_help)
        sizer.Add(self.btn_nose_help, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        row += 1
        col = 0
        self.btn_load_tests_from_files = Button(self, label="Load tests from files")
        self.btn_load_tests_from_files.Bind(EVT_BUTTON, self.__load_tests_from_files)
        sizer.Add(
            self.btn_load_tests_from_files, pos=(row, col), flag=FLAG_ALL_AND_EXPAND
        )

        col += 1
        self.btn_load_tests_from_dir = Button(self, label="Load tests from directory")
        self.btn_load_tests_from_dir.Bind(EVT_BUTTON, self.__load_tests_from_directory)
        sizer.Add(
            self.btn_load_tests_from_dir, pos=(row, col), flag=FLAG_ALL_AND_EXPAND
        )

        col += 1
        dummy_label = StaticText(self)
        sizer.Add(dummy_label, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.cb_browser = Choice(self, choices=Browser.get_supported_browsers())
        self.cb_browser.Select(0)
        sizer.Add(self.cb_browser, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        col += 1
        self.btn_run = Button(self, label="Run test cases")
        self.btn_run.Bind(EVT_BUTTON, self.__run_tests)
        sizer.Add(self.btn_run, pos=(row, col), flag=FLAG_ALL_AND_EXPAND)

        row += 1
        col = 0
        window = SplitterWindow(self, style=SP_3D | SP_LIVE_UPDATE)
        self.tree_ctrl = CustomTreeCtrl(
            window,
            style=TR_SINGLE
            | TR_HAS_BUTTONS
            | TR_AUTO_CHECK_CHILD
            | TR_AUTO_CHECK_PARENT
            | TR_AUTO_TOGGLE_CHILD,
        )
        self.tree_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        self.tree_ctrl.SetForegroundColour(self.GetForegroundColour())
        self.tree_ctrl.Bind(EVT_TREE_ITEM_CHECKED, self.__on_tree_check)

        self.txt_ctrl = TextCtrl(
            window, style=TE_MULTILINE | TE_READONLY | HSCROLL | VSCROLL
        )
        font_size = self.txt_ctrl.GetFont().GetPointSize()
        self.txt_ctrl.SetFont(Font(font_size, FONTFAMILY_TELETYPE, NORMAL, NORMAL))

        window.SplitVertically(self.tree_ctrl, self.txt_ctrl)
        sizer.Add(window, pos=(row, col), span=(1, 5), flag=FLAG_ALL_AND_EXPAND)

        sizer.AddGrowableCol(2, 1)
        sizer.AddGrowableRow(row, 1)
        self.SetSizerAndFit(sizer)

    def __on_show_help(self, evt):
        # TODO: simplify
        text = check_output(
            [
                os.sys.executable,
                os.path.abspath(
                    os.path.join(
                        os.path.dirname(__file__), "../scripts/easelenium_cli.py"
                    )
                ),
                "--help",
            ]
        )
        DialogWithText(self, "Help for nosetests", text).ShowModal()

    def __on_select_file(self, evt):
        folder = self.__get_safe_path_from_root_folder(RootFolder.REPORTS)
        obj = evt.GetEventObject()
        txt_ctrl = None
        if obj == self.btn_select_html:
            wildcard = "*.html"
            txt_ctrl = self.txt_html_report
        elif obj == self.btn_select_xml:
            wildcard = "*.xml"
            txt_ctrl = self.txt_xml_report
        else:
            wildcard = "*.*"
        dialog = FileDialog(
            self,
            defaultDir=folder,
            style=FD_SAVE | FD_OVERWRITE_PROMPT,
            wildcard=wildcard,
        )
        if dialog.ShowModal() == ID_OK and txt_ctrl:
            txt_ctrl.SetValue(dialog.GetPath())

    def __on_check(self, evt):
        cb_obj = evt.GetEventObject()
        checkboxes_and_txt_ctrls = {
            self.cb_html_output: self.txt_html_report,
            self.cb_options: self.txt_options,
            self.cb_xml_output: self.txt_xml_report,
        }
        checkboxes_and_btns = {
            self.cb_html_output: self.btn_select_html,
            self.cb_xml_output: self.btn_select_xml,
        }
        txt_ctrl = checkboxes_and_txt_ctrls[cb_obj]
        btn = checkboxes_and_btns.get(cb_obj)
        if txt_ctrl.IsEnabled():
            txt_ctrl.Disable()
            if btn:
                btn.Disable()
        else:
            txt_ctrl.Enable()
            if btn:
                btn.Enable()

    def __on_tree_check(self, evt):
        # styles doesn't work:
        # TR_AUTO_CHECK_CHILD | TR_AUTO_CHECK_PARENT | TR_AUTO_TOGGLE_CHILD
        # TODO: fix if all children are checked
        # then one child is unchecked - parent is checked
        item = evt.GetItem()
        checked = item.IsChecked()
        parent = item.GetParent()

        def all_children_are_checked(_parent):
            states = [child.IsChecked() for child in _parent.GetChildren()]
            uniq_states = list(set(states))
            return len(uniq_states) == 1 and uniq_states[0] is True

        self.tree_ctrl.AutoCheckChild(item, checked)
        if parent:
            self.tree_ctrl.AutoCheckParent(item, all_children_are_checked(parent))

    def __get_safe_path_from_root_folder(self, subfolder=None):
        folder = self.GetTopLevelParent().get_root_folder()
        if subfolder and folder:
            path_for_subfolder = os.path.join(folder, subfolder)
            if os.path.exists(path_for_subfolder):
                return path_for_subfolder

        return folder if folder else "."

    def __load_tests_to_tree(self, file_paths=None, dir_path=None):
        if file_paths:
            python_files = file_paths
        elif dir_path:
            python_files = [f for f in get_list_of_files(dir_path, True)]
        else:
            python_files = []

        python_files = [
            f
            for f in python_files
            if "test" in os.path.basename(f) and os.path.splitext(f)[-1] == ".py"
        ]
        if len(python_files) > 0:
            syspath = list(os.sys.path)
            try:
                root_folder = self.__get_safe_path_from_root_folder()

                if root_folder not in os.sys.path:
                    os.sys.path.append(root_folder)

                checkbox_type = 1
                self.tree_ctrl.DeleteAllItems()
                root = self.tree_ctrl.AddRoot("All test cases", checkbox_type)

                for python_file in python_files:
                    top_item = self.tree_ctrl.AppendItem(
                        root, os.path.abspath(python_file), checkbox_type
                    )

                    parsed_classes = ParsedClass.get_parsed_classes(python_file)
                    for parsed_class in parsed_classes:
                        item = self.tree_ctrl.AppendItem(
                            top_item, parsed_class.name, checkbox_type
                        )

                        test_methods = [
                            k
                            for k in parsed_class.methods.keys()
                            if k.startswith("test_")
                        ]
                        for tc_name in test_methods:
                            self.tree_ctrl.AppendItem(item, tc_name, checkbox_type)

                self.tree_ctrl.ExpandAll()
            except Exception:
                show_error_dialog(self, traceback.format_exc(), "Cannot add test cases")
            finally:
                os.sys.path = syspath

    def __load_tests_from_directory(self, evt):
        folder = self.__get_safe_path_from_root_folder(RootFolder.TESTS_FOLDER)
        if folder:
            dialog = DirDialog(self, defaultPath=folder, style=DD_DIR_MUST_EXIST)
            if dialog.ShowModal() == ID_OK:
                self.__load_tests_to_tree(dir_path=dialog.GetPath())
        else:
            show_dialog_path_doesnt_exist(self, folder)

    def __load_tests_from_files(self, evt):
        folder = self.__get_safe_path_from_root_folder(RootFolder.TESTS_FOLDER)
        if folder:
            dialog = FileDialog(
                self,
                defaultDir=folder,
                style=FD_OPEN | FD_FILE_MUST_EXIST | FD_MULTIPLE,
                wildcard="*.py",
            )
            if dialog.ShowModal() == ID_OK:
                self.__load_tests_to_tree(file_paths=dialog.GetPaths())
        else:
            show_dialog_path_doesnt_exist(self, folder)

    def __get_command(self):
        root = self.tree_ctrl.GetRootItem()
        tests = []
        for _file in root.GetChildren():
            for _class in _file.GetChildren():
                for test_case in _class.GetChildren():
                    if test_case.IsChecked():
                        # TODO: fix for files that contain spaces
                        test_path = _file.GetText()
                        test_class = _class.GetText()
                        test_method = test_case.GetText()
                        tests.append(f"{test_path}")
                        tests.append("-k")
                        tests.append(f"{test_class} and {test_method}")

        if not tests:
            return

        args = [
            f"--rootdir={self.__get_safe_path_from_root_folder()}",
        ]

        use_html_report = (
            self.cb_html_output.IsChecked() and len(self.txt_html_report.GetValue()) > 0
        )
        if use_html_report:
            report_path = self.txt_html_report.GetValue()
            args.append(f"--html={report_path}")

        use_xml_report = (
            self.cb_xml_output.IsChecked() and len(self.txt_xml_report.GetValue()) > 0
        )
        if use_xml_report:
            report_path = self.txt_xml_report.GetValue()
            args.append(f"--junitxml={report_path}")

        use_options = (
            self.cb_options.IsChecked() and len(self.txt_options.GetValue()) > 0
        )
        if use_options:
            report_path = self.txt_options.GetValue()
            args.append(report_path)

        nose_cmd = ["easelenium_cli.py"] + args + tests
        return nose_cmd

    def __run_tests(self, evt):
        # TODO: do not run if root folder is not selected
        self.txt_ctrl.Clear()

        dialog = InfiniteProgressBarDialog(
            self,
            "Running test cases",
            "Running selected test cases... Please wait...",
        )

        def wrap_func():
            stdout = os.sys.stdout
            stderr = os.sys.stderr
            redirected = RedirectText(self.txt_ctrl)
            os.sys.stdout = redirected
            os.sys.stderr = redirected
            try:
                cmd = self.__get_command()
                browser_name = self.cb_browser.GetStringSelection()
                Browser.DEFAULT_BROWSER = browser_name
                report_folder = self.__get_safe_path_from_root_folder(
                    RootFolder.REPORTS
                )
                BaseTest.FAILED_SCREENSHOT_FOLDER = report_folder

                easelenium_cmd = " ".join(cmd).replace(
                    "easelenium_cli.py",
                    "easelenium_cli.py --browser " + browser_name,
                )
                print(f"Executing command:\n{easelenium_cmd}")

                main(cmd[1:])
            finally:
                dialog.close_event.set()
                os.sys.stdout = stdout
                os.sys.stderr = stderr

        run_in_separate_thread(wrap_func)
        dialog.ShowModal()



from wx import EVT_MENU, Menu, MenuItem, NewId


class MenuItemData(object):
    def __init__(self, text, func):
        self.text = text
        self.func = func
        self.id = NewId()

    def __str__(self):
        return f"MenuItemData({self.__dict__})"

    def __repr__(self):
        return str(self)


class ContextMenu(Menu):
    SEPARATOR_TEXT = "------"

    def __init__(self, data):
        Menu.__init__(self)

        self.__id_and_item_data = self.__get_ids_and_item_data(data)
        self.__text_and_item_data = dict(
            [(d.text, d) for d in self.__id_and_item_data.values()]
        )

        self.__create_menu(data)
        self._bind_evt_menu(self.__on_menu_click)

    def _bind_evt_menu(self, function):
        self.Bind(EVT_MENU, function)
        for menu_item in self.GetMenuItems():
            submenu = menu_item.GetSubMenu()
            if submenu:
                submenu.Bind(EVT_MENU, function)

    def __get_ids_and_item_data(self, data):
        _ids_and_data = dict()
        for text, func in data:
            if type(func) in (list, tuple):
                for _id, _item_data in self.__get_ids_and_item_data(func).items():
                    _ids_and_data[_id] = _item_data
            item_data = MenuItemData(text, func)
            _ids_and_data[item_data.id] = item_data
        return _ids_and_data

    def __create_menu(self, data):
        for text, func in data:
            if type(func) in (list, tuple):
                submenu = Menu()
                for _text, _ in func:
                    _id = self.__text_and_item_data[_text].id
                    sitem = MenuItem(submenu, _id, _text)
                    submenu.Append(sitem)
                self.AppendSubMenu(submenu, text)
            else:
                if text == self.SEPARATOR_TEXT:
                    self.AppendSeparator()
                else:
                    data = self.__text_and_item_data[text]
                    item = MenuItem(self, data.id, text)
                    self.Append(item)

    def _get_menu_item_data(self, menu_item_id):
        return self.__id_and_item_data.get(menu_item_id)

    def _get_function(self, menu_item_id):
        menu_item_data = self._get_menu_item_data(menu_item_id)
        return menu_item_data.func if menu_item_data else None

    def __on_menu_click(self, evt):
        _id = evt.GetId()
        func = self._get_function(_id)
        if func:
            func()



# coding=utf8
import os
import traceback
from tempfile import gettempdir

from selenium import webdriver
from selenium.common.exceptions import (
    NoSuchElementException,
    TimeoutException,
    WebDriverException,
)
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.select import Select
from selenium.webdriver.support.wait import WebDriverWait

from easelenium.mouse import Mouse
from easelenium.utils import get_random_value, get_timestamp, is_windows


def browser_decorator(
    browser_name=None, timeout=5, logger=None, headless=False, webdriver_kwargs=None
):
    def func_decorator(func):
        def wrapper(*args, **kwargs):
            browser = None
            return_value = None
            try:
                browser = Browser(
                    browser_name=browser_name,
                    logger=logger,
                    timeout=timeout,
                    headless=headless,
                    webdriver_kwargs=webdriver_kwargs,
                )

                kwargs["browser"] = browser
                value = func(*args, **kwargs)
                return_value = value
            except Exception:
                try:
                    if browser:
                        browser.save_screenshot()
                except Exception:
                    pass
                traceback.print_exc()
            finally:
                if browser:
                    browser.quit()

            return return_value

        return wrapper

    return func_decorator


class Browser(object):
    FF = "ff"
    FF_HEADLESS = "ff_headless"
    GC = "gc"
    GC_HEADLESS = "gc_headless"
    IE = "ie"
    OP = "op"
    PHANTOMJS = "phantomjs"
    DEFAULT_BROWSER = None

    __BROWSERS = [FF, FF_HEADLESS, GC, GC_HEADLESS, IE, OP, PHANTOMJS]

    __DRIVERS_AND_CONSTUCTORS = driver_and_constructor = {
        FF: ("geckodriver", webdriver.Firefox),
        FF_HEADLESS: ("geckodriver", webdriver.Firefox),
        IE: ("IEDriverServer", webdriver.Ie),
        GC: ("chromedriver", webdriver.Chrome),
        GC_HEADLESS: ("chromedriver", webdriver.Chrome),
        OP: ("operadriver", webdriver.Opera),
        PHANTOMJS: ("phantomjs", webdriver.PhantomJS),
    }
    __LOCATOR_MAPPINGS = {
        "by_name": By.NAME,
        "by_id": By.ID,
        "by_xpath": By.XPATH,
        "by_link": By.LINK_TEXT,
        "by_partial_link": By.PARTIAL_LINK_TEXT,
        "by_tag": By.TAG_NAME,
        "by_class": By.CLASS_NAME,
        "by_css": By.CSS_SELECTOR,
    }

    def __init__(
        self,
        browser_name=None,
        logger=None,
        timeout=5,
        headless=False,
        webdriver_kwargs=None,
    ):
        if webdriver_kwargs is None:
            webdriver_kwargs = {}

        self.__browser_name = self.DEFAULT_BROWSER or browser_name or self.FF

        self.logger = logger
        self.__timeout = timeout

        headless = headless or "headless" in self.__browser_name
        if self.is_gc():
            self.__set_chrome_kwargs(headless, webdriver_kwargs)
        elif self.is_ff():
            self.__set_firefox_kwargs(headless, webdriver_kwargs)

        self._driver = self.__create_driver(self.__browser_name, webdriver_kwargs)
        self.__screenshot_path = os.path.join(gettempdir(), "easelenium_screenshots")

        if not os.path.exists(self.__screenshot_path):
            os.makedirs(self.__screenshot_path)

        self.mouse = Mouse(self)

    def __set_chrome_kwargs(self, headless, webdriver_kwargs):
        options = webdriver_kwargs.get("options", webdriver.ChromeOptions())
        if headless:
            options.add_argument("--headless")
        webdriver_kwargs["options"] = options

    def __set_firefox_kwargs(self, headless, webdriver_kwargs):
        options = webdriver_kwargs.get("options", webdriver.FirefoxOptions())
        if headless:
            options.add_argument("--headless")
        webdriver_kwargs["options"] = options

    @classmethod
    def supports(cls, browser_name):
        return cls._find_driver_path(browser_name) is not None

    @classmethod
    def _find_driver_path(cls, browser_name):
        assert browser_name in cls.__BROWSERS

        driver_filename = cls.__DRIVERS_AND_CONSTUCTORS[browser_name][0]
        if is_windows():
            driver_filename += ".exe"

        possible_path = os.path.join(os.path.expanduser("~"), driver_filename)
        if os.path.exists(possible_path):
            return possible_path
        else:
            for folder in os.getenv("PATH", "").split(os.pathsep):
                possible_path = os.path.join(folder, driver_filename)
                if os.path.exists(possible_path):
                    return possible_path

        return None

    @classmethod
    def get_supported_browsers(cls):
        return cls.__BROWSERS

    def get_browser_initials(self):
        return self.__browser_name

    def __create_driver(self, name, webdriver_kwargs):
        driver_filename_and_constructor = self.__DRIVERS_AND_CONSTUCTORS.get(name, None)

        if driver_filename_and_constructor is None:
            browsers = "', '".join(self.__BROWSERS)
            raise ValueError(
                f"Unsupported browser '{name}', supported browsers: ['{browsers}']"
            )

        driver_filename, constructor = driver_filename_and_constructor

        driver_path = webdriver_kwargs.get("executable_path") or self._find_driver_path(
            name
        )
        if driver_path is None:
            raise FileNotFoundError(f"Failed to find {driver_filename}")

        webdriver_kwargs["executable_path"] = driver_path

        driver = constructor(**webdriver_kwargs)

        if driver and not self.is_gc() and not self.is_op():
            driver.maximize_window()

        return driver

    def get_by_query(
        self,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        # TODO: move outside
        for locator, value in (
            ("by_id", by_id),
            ("by_xpath", by_xpath),
            ("by_tag", by_tag),
            ("by_name", by_name),
            ("by_css", by_css),
            ("by_class", by_class),
            ("by_link", by_link),
            ("by_partial_link", by_partial_link),
        ):
            if value is not None:
                return (self.__LOCATOR_MAPPINGS[locator], value)
        else:
            raise ValueError("Failed to find element")

    def _get_element(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        assert (
            element is not None
            or by_id is not None
            or by_xpath is not None
            or by_link is not None
            or by_partial_link is not None
            or by_name is not None
            or by_tag is not None
            or by_css is not None
            or by_class is not None
        ), "'element' or 'by_*' not specified"

        if element:
            return element
        else:
            return self.get_by_query(
                by_id=by_id,
                by_xpath=by_xpath,
                by_link=by_link,
                by_partial_link=by_partial_link,
                by_name=by_name,
                by_tag=by_tag,
                by_css=by_css,
                by_class=by_class,
            )

    def __get_webelements(
        self,
        element=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        if isinstance(element, WebElement):
            return [element]
        elif type(element) in [list, tuple] and len(element) == 2:
            if parent:
                return parent.find_elements(*element)
            else:
                return self._driver.find_elements(*element)
        else:
            raise Exception("Unsupported element - %s" % str(element))

    def __get_by_and_locator(self, element):
        # TODO: remove not needed
        if (type(element) == list or type(element) == tuple) and len(element) == 2:
            by, locator = element
            return by, locator
        else:
            return None

    def _to_string(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        # TODO: move outside
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        by_and_locator = self.__get_by_and_locator(element)
        if by_and_locator:
            return "Element {By: '%s', value: '%s'}" % (
                by_and_locator[0],
                by_and_locator[1],
            )
        else:
            string = "tag_name: '%s'" % element.tag_name
            _id = element.get_attribute("id")
            if _id:
                string += ", id: '%s'" % _id
            class_name = element.get_attribute("class")
            if class_name:
                string += ", class: '%s'" % class_name
            text = element.text
            if text:
                string += ", text: '%s'" % text
            value = element.get_attribute("value")
            if value:
                string += ", value: '%s'" % value
            name = element.get_attribute("name")
            if name and element.tag_name in ["frame", "iframe"]:
                string += ", name: '%s'" % name
            return "Element {%s}" % string

    def is_ff(self):
        return self.__browser_name.startswith(Browser.FF)

    def is_ie(self):
        return self.__browser_name == Browser.IE

    def is_gc(self):
        return self.__browser_name.startswith(Browser.GC)

    def is_op(self):
        return self.__browser_name == Browser.OP

    def _safe_log(self, *args):
        if self.logger:
            args = [
                self._to_string(arg) if isinstance(arg, WebElement) else arg
                for arg in args
            ]
            self.logger.info(*args)

    """
        WebElement's wrapped functions
    """

    def type(
        self,
        element=None,
        text=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

        self.wait_for_visible(
            element=element,
        )
        element = self.find_element(
            element=element,
        )
        try:
            element.clear()
        except WebDriverException as e:
            if "Element must be user-editable in order to clear it." != e.msg:
                raise e

        self._safe_log("Typing '%s' at '%s'", text, element)

        element.send_keys(text)

    def click(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )
        element = self.find_element(
            element=element,
        )

        self._safe_log("Clicking at '%s'", element)

        element.click()

    def get_parent(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self.find_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        return self.find_descendant(parent=element, by_xpath="./..")

    def get_text(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )
        element = self.find_element(
            element=element,
        )
        text = element.text

        self._safe_log("Getting text from '%s' -> '%s'", element, text)

        return text

    def get_attribute(
        self,
        element=None,
        attr=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        assert attr is not None, "attr is not specified"
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )
        element = self.find_elements(
            element=element,
        )[0]
        value = element.get_attribute(attr)

        self._safe_log(f"Getting attribute {attr} from {element} -> {value}")

        return value

    def get_tag_name(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return self.find_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        ).tag_name

    def get_id(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return self.get_attribute(
            element=element,
            attr="id",
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def get_class(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return self.get_attribute(
            element=element,
            attr="class",
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def get_value(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return self.get_attribute(
            element=element,
            attr="value",
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def get_location(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        """Return tuple like (x, y)."""
        location = self.find_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        ).location

        self._safe_log(f"Getting location from {element} -> {location}")

        return int(location["x"]), int(location["y"])

    def get_dimensions(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        """Return tuple like (width, height)."""
        size = self.find_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        ).size

        self._safe_log(f"Getting dimensions from {element} -> {size}")

        return size["width"], size["height"]

    """
        Dropdown list related methods
    """

    def get_selected_value_from_dropdown(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )

        element = self.find_element(
            element=element,
        )
        value = Select(element).first_selected_option.get_attribute("value")

        self._safe_log("Getting selected value from '%s' -> '%s'", element, value)

        return value

    def get_selected_text_from_dropdown(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )

        element = self.find_element(
            element=element,
        )

        text = Select(element).first_selected_option.text

        self._safe_log("Getting selected text from '%s' -> '%s'", element, text)

        return text

    def select_option_by_value_from_dropdown(
        self,
        element=None,
        value=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )

        element = self.find_element(
            element=element,
        )
        select = Select(element)
        assert value is not None, "value not specified"

        self._safe_log(f"Selecting by value {value} from {element}")

        select.select_by_value(value)

    def select_option_by_text_from_dropdown(
        self,
        element=None,
        text=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )

        element = self.find_element(
            element=element,
        )
        select = Select(element)

        self._safe_log(f"Selecting by text {text} from {element}")

        select.select_by_visible_text(text)

    def select_option_by_index_from_dropdown(
        self,
        element=None,
        index=0,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

        self.wait_for_visible(
            element=element,
        )

        element = self.find_element(
            element=element,
        )
        select = Select(element)

        self._safe_log(f"Selecting by index {index} from {element}")

        select.select_by_index(index)

    def select_random_option_from_dropdown(
        self,
        element=None,
        texts_to_skip=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )
        texts_to_skip = texts_to_skip or []

        options = self.get_texts_from_dropdown(
            element=element,
        )
        option_to_select = get_random_value(options, *texts_to_skip)

        self.select_option_by_text_from_dropdown(
            element=element,
            text=option_to_select,
        )

    def get_texts_from_dropdown(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )

        texts = []
        element = self.find_element(
            element=element,
        )
        for option in Select(element).options:
            texts.append(option.text)

        self._safe_log("Getting texts from '%s' -> '%s'", element, str(texts))

        return texts

    def get_values_from_dropdown(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        self.wait_for_visible(
            element=element,
        )

        values = []
        element = self.find_element(
            element=element,
        )
        for option in Select(element).options:
            values.append(option.get_attribute("value"))

        self._safe_log("Getting values from '%s' -> '%s'", element, str(values))

        return values

    """
        WebDriver's wrapped functions
    """

    def get_action_chains(self):
        return ActionChains(self._driver)

    def open(self, url):
        self.get(url)

    def get(self, url):
        self._driver.get(url)

    def execute_js(self, js_script, *args):
        return self._driver.execute_script(js_script, *args)

    def find_element(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return self.find_descendant(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def find_descendant(
        self,
        element=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        found_elements = self.find_descendants(
            element=element,
            parent=parent,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        if len(found_elements) == 0:
            raise NoSuchElementException(
                "Didn't find any elements for selector - %s" % str(element)
            )
        else:
            return found_elements[0]

    def find_elements(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        elements = self.__get_webelements(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        if type(elements) == list:
            return elements
        else:
            return [elements]

    def find_descendants(
        self,
        parent=None,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return self.__get_webelements(
            parent=parent,
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

    def wait_for_text_is_changed(
        self, element=None, old_text=None, msg=None, timeout=None
    ):
        if not timeout:
            timeout = self.__timeout
        if not msg:
            msg = "%s text was not changed for %s seconds" % (element, timeout)

        self.webdriver_wait(
            lambda driver: old_text != self.get_text(element), msg, timeout
        )

    def wait_for_attribute_is_changed(
        self, element=None, attr=None, old_value=None, msg=None, timeout=None
    ):
        if not timeout:
            timeout = self.__timeout
        if not msg:
            msg = "%s attribute was not changed for %s seconds" % (element, timeout)

        self.webdriver_wait(
            lambda driver: old_value != self.get_attribute(element, attr, False),
            msg,
            timeout,
        )

    def wait_for_visible(
        self,
        element=None,
        msg=None,
        timeout=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        if not timeout:
            timeout = self.__timeout
        if not msg:
            msg = "%s is not visible for %s seconds" % (element, timeout)
        self.webdriver_wait(
            lambda driver: self.is_visible(element=element, parent=parent),
            msg,
            timeout,
        )

    def wait_for_not_visible(
        self,
        element=None,
        msg=None,
        timeout=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self._get_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )
        if not timeout:
            timeout = self.__timeout
        if not msg:
            msg = "%s is visible for %s seconds" % (element, timeout)

        self.webdriver_wait(
            lambda driver: not self.is_visible(element=element, parent=parent),
            msg,
            timeout,
        )

    def wait_for_present(
        self,
        element=None,
        msg=None,
        timeout=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        if not timeout:
            timeout = self.__timeout
            msg = "%s is not present for %s seconds" % (element, timeout)

        self.webdriver_wait(lambda driver: self.is_present(element), msg, timeout)

    def wait_for_not_present(
        self,
        element=None,
        msg=None,
        timeout=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        if not timeout:
            timeout = self.__timeout
            msg = "%s is present for %s seconds" % (element, timeout)

        self.webdriver_wait(lambda driver: not self.is_present(element), msg, timeout)

    def is_visible(
        self,
        element=None,
        parent=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        try:
            if parent:
                elements = self.find_descendants(
                    element=element,
                    parent=parent,
                    by_id=by_id,
                    by_xpath=by_xpath,
                    by_link=by_link,
                    by_partial_link=by_partial_link,
                    by_name=by_name,
                    by_tag=by_tag,
                    by_css=by_css,
                    by_class=by_class,
                )
            else:
                elements = self.find_elements(
                    element=element,
                    by_id=by_id,
                    by_xpath=by_xpath,
                    by_link=by_link,
                    by_partial_link=by_partial_link,
                    by_name=by_name,
                    by_tag=by_tag,
                    by_css=by_css,
                    by_class=by_class,
                )
            return len(elements) > 0 and elements[0].is_displayed()
        except WebDriverException:
            return False

    def is_present(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return (
            len(
                self.find_elements(
                    element=element,
                    by_id=by_id,
                    by_xpath=by_xpath,
                    by_link=by_link,
                    by_partial_link=by_partial_link,
                    by_name=by_name,
                    by_tag=by_tag,
                    by_css=by_css,
                    by_class=by_class,
                )
            )
            > 0
        )

    def get_screenshot_as_png(self):
        return self._driver.get_screenshot_as_png()

    def save_screenshot(self, saving_dir=None, filename=None):
        if not saving_dir:
            saving_dir = self.__screenshot_path
        if not filename:
            filename = get_timestamp() + ".png"
        path_to_file = os.path.abspath(os.path.join(saving_dir, filename))

        self._safe_log("Saving screenshot to '%s'", path_to_file)

        self._driver.save_screenshot(path_to_file)
        return path_to_file

    def get_elements_count(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        return len(
            self.find_elements(
                element=element,
                by_id=by_id,
                by_xpath=by_xpath,
                by_link=by_link,
                by_partial_link=by_partial_link,
                by_name=by_name,
                by_tag=by_tag,
                by_css=by_css,
                by_class=by_class,
            )
        )

    def switch_to_frame(
        self,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        element = self.find_element(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

        self._safe_log("Switching to '%s' frame", element)

        self._driver.switch_to.frame(element)

    def switch_to_new_window(
        self,
        function,
        element=None,
        by_id=None,
        by_xpath=None,
        by_link=None,
        by_partial_link=None,
        by_name=None,
        by_tag=None,
        by_css=None,
        by_class=None,
    ):
        initial_handles = self._driver.window_handles

        function(
            element=element,
            by_id=by_id,
            by_xpath=by_xpath,
            by_link=by_link,
            by_partial_link=by_partial_link,
            by_name=by_name,
            by_tag=by_tag,
            by_css=by_css,
            by_class=by_class,
        )

        self.webdriver_wait(
            lambda driver: len(initial_handles) < len(driver.window_handles)
        )
        new_handles = self._driver.window_handles
        for handle in initial_handles:
            new_handles.remove(handle)

        self._driver.switch_to.window(new_handles[0])

        self._safe_log("Switching to '%s' window", self._driver.title)

    def switch_to_default_content(self):
        self._safe_log("Switching to default content")

        self._driver.switch_to.default_content()

    def close_current_window_and_focus_to_previous_one(self):
        handles = self._driver.window_handles
        self.close()
        self._driver.switch_to.window(handles[-2])

    def get_page_source(self):
        return self._driver.page_source

    def get_title(self):
        return self._driver.title

    def get_current_url(self):
        return self._driver.current_url

    def get_current_frame_url(self):
        return self.execute_js("return document.location.href")

    def go_back(self):
        self._driver.back()

    def delete_all_cookies(self):
        self._driver.delete_all_cookies()

    def alert_accept(self):
        self._safe_log("Clicking Accept/OK in alert box")

        self._driver.switch_to.alert.accept()

    def alert_dismiss(self):
        self._safe_log("Clicking Dismiss/Cancel in alert box")

        self._driver.switch_to.alert.dismiss()

    def refresh_page(self):
        self._driver.refresh()

    def webdriver_wait(self, function, msg="", timeout=None):
        if not timeout:
            timeout = self.__timeout
        try:
            WebDriverWait(self._driver, timeout).until(function, msg)
        except Exception:
            raise TimeoutException(msg)

    def close(self):
        self._driver.close()

    def quit(self):
        self._driver.quit()
